from contextlib import nullcontext
from typing import Any, ContextManager, Optional, Type

import pytest
import yaml
from pydantic import ValidationError

from elabftwcontrol.core.manifests import (
    BaseMetaField,
    DependencyGraph,
    ElabObjManifests,
    ExperimentManifest,
    ExperimentSpecManifest,
    ExperimentSpecManifestSimplifiedMetadata,
    ExperimentTemplateManifest,
    ExperimentTemplateSpecManifest,
    ExtraFieldsManifest,
    ExtraFieldsManifestComplex,
    FieldTypeEnum,
    ItemManifest,
    ItemSpecManifest,
    ItemSpecManifestSimplifiedMetadata,
    ItemsTypeManifest,
    ItemsTypeSpecManifest,
    ManifestIndex,
    MetadataCheckboxFieldManifest,
    MetadataCheckboxFieldOptions,
    MetadataDateFieldManifest,
    MetadataDatetimeFieldManifest,
    MetadataEmailFieldManifest,
    MetadataExperimentsLinkFieldManifest,
    MetadataFieldManifest,
    MetadataGroupManifest,
    MetadataItemsLinkFieldManifest,
    MetadataManifestConfig,
    MetadataNumberFieldManifest,
    MetadataRadioFieldManifest,
    MetadataSelectFieldManifest,
    MetadataTextFieldManifest,
    MetadataTimeFieldManifest,
    MetadataURLFieldManifest,
)
from elabftwcontrol.core.manifests import NameNode as Node
from elabftwcontrol.core.manifests import (
    ObjectTypes,
    SimpleExtraFieldsManifest,
    _ValueAndUnit,
)


@pytest.mark.parametrize(
    ("input", "field_type", "expected", "expectation"),
    (
        (
            {"group": 1},
            BaseMetaField,
            None,
            pytest.raises(ValidationError),
        ),
        (
            {"name": "fieldname", "group": "groupname"},
            BaseMetaField,
            BaseMetaField(
                name="fieldname",
                group="groupname",
            ),
            nullcontext(),
        ),
        (
            {"name": "fieldname", "non_existing": 1},
            BaseMetaField,
            None,
            pytest.raises(ValidationError),
        ),
        (
            {"type": "checkbox", "name": "fieldname", "value": "on"},
            MetadataCheckboxFieldManifest,
            MetadataCheckboxFieldManifest(
                type=FieldTypeEnum.checkbox,
                name="fieldname",
                value=MetadataCheckboxFieldOptions("on"),
            ),
            nullcontext(),
        ),
        (
            {"type": "checkbox", "name": "fieldname", "value": "off"},
            MetadataCheckboxFieldManifest,
            MetadataCheckboxFieldManifest(
                type=FieldTypeEnum.checkbox,
                name="fieldname",
                value=MetadataCheckboxFieldOptions("off"),
            ),
            nullcontext(),
        ),
        (
            {"type": "checkbox", "name": "fieldname", "value": "something_else"},
            MetadataCheckboxFieldManifest,
            None,
            pytest.raises(ValueError),
        ),
        (
            {"type": "wrongtype", "name": "fieldname", "value": "on"},
            MetadataCheckboxFieldManifest,
            None,
            pytest.raises(ValueError),
        ),
        (
            {
                "type": "radio",
                "name": "fieldname",
                "options": ["option 1", "option 2"],
                "value": "option 1",
            },
            MetadataRadioFieldManifest,
            MetadataRadioFieldManifest(
                type=FieldTypeEnum.radio,
                name="fieldname",
                value="option 1",
                options=["option 1", "option 2"],
            ),
            nullcontext(),
        ),
        (
            {
                "type": "radio",
                "name": "fieldname",
                "options": ["option 1", "option 2"],
                "value": "option 3",
            },
            MetadataRadioFieldManifest,
            None,
            pytest.raises(ValueError),
        ),
        (
            {
                "type": "radio",
                "name": "fieldname",
                "value": "option 3",
            },
            MetadataRadioFieldManifest,
            None,
            pytest.raises(ValueError),
        ),
        (
            {
                "type": "radio",
                "name": "fieldname",
                "options": ["option 1", "option 2"],
            },
            MetadataRadioFieldManifest,
            MetadataRadioFieldManifest(
                type=FieldTypeEnum.radio,
                name="fieldname",
                value="",
                options=["option 1", "option 2"],
            ),
            nullcontext(),
        ),
        (
            {
                "type": "select",
                "name": "fieldname",
                "options": ["option 1", "option 2"],
                "value": "option 1",
            },
            MetadataSelectFieldManifest,
            MetadataSelectFieldManifest(
                type=FieldTypeEnum.select,
                name="fieldname",
                value="option 1",
                options=["option 1", "option 2"],
            ),
            nullcontext(),
        ),
        (
            {
                "type": "select",
                "name": "fieldname",
                "options": ["option 1", "option 2"],
                "value": ["option 1"],
                "allow_multi_values": True,
            },
            MetadataSelectFieldManifest,
            MetadataSelectFieldManifest(
                type=FieldTypeEnum.select,
                name="fieldname",
                value=["option 1"],
                options=["option 1", "option 2"],
                allow_multi_values=True,
            ),
            nullcontext(),
        ),
        (
            {
                "type": "select",
                "name": "fieldname",
                "options": ["option 1", "option 2"],
                "value": ["option 1", "option 2"],
                "allow_multi_values": True,
            },
            MetadataSelectFieldManifest,
            MetadataSelectFieldManifest(
                type=FieldTypeEnum.select,
                name="fieldname",
                value=["option 1", "option 2"],
                options=["option 1", "option 2"],
                allow_multi_values=True,
            ),
            nullcontext(),
        ),
        (
            {
                "type": "select",
                "name": "fieldname",
                "options": ["option 1", "option 2"],
                "value": "option 3",
            },
            MetadataSelectFieldManifest,
            None,
            pytest.raises(ValueError),
        ),
        (
            {
                "type": "select",
                "name": "fieldname",
                "options": ["option 1", "option 2"],
                "value": "option 1",
                "allow_multi_values": True,
            },
            MetadataSelectFieldManifest,
            None,
            pytest.raises(ValueError),
        ),
        (
            {
                "type": "select",
                "name": "fieldname",
                "options": ["option 1", "option 2"],
                "value": ["option 3"],
                "allow_multi_values": True,
            },
            MetadataSelectFieldManifest,
            None,
            pytest.raises(ValueError),
        ),
        (
            {
                "type": "select",
                "name": "fieldname",
                "value": "option 3",
            },
            MetadataSelectFieldManifest,
            None,
            pytest.raises(ValueError),
        ),
        (
            {
                "type": "select",
                "name": "fieldname",
                "options": ["option 1", "option 2"],
            },
            MetadataSelectFieldManifest,
            MetadataSelectFieldManifest(
                type=FieldTypeEnum.select,
                name="fieldname",
                value="",
                options=["option 1", "option 2"],
            ),
            nullcontext(),
        ),
        (
            {
                "type": "number",
                "name": "fieldname",
                "value": "1234.1",
            },
            MetadataNumberFieldManifest,
            MetadataNumberFieldManifest(
                type=FieldTypeEnum.number,
                name="fieldname",
                value="1234.1",
            ),
            nullcontext(),
        ),
        (
            {
                "type": "number",
                "name": "fieldname",
            },
            MetadataNumberFieldManifest,
            MetadataNumberFieldManifest(
                type=FieldTypeEnum.number,
                name="fieldname",
            ),
            nullcontext(),
        ),
        (
            {
                "type": "number",
                "name": "fieldname",
                "value": "1234.1",
                "unit": "kg",
                "units": ["g", "kg"],
            },
            MetadataNumberFieldManifest,
            MetadataNumberFieldManifest(
                type=FieldTypeEnum.number,
                name="fieldname",
                value="1234.1",
                unit="kg",
                units=["g", "kg"],
            ),
            nullcontext(),
        ),
        (
            {
                "type": "number",
                "name": "fieldname",
                "value": "1234.1",
                "units": ["g", "kg"],
            },
            MetadataNumberFieldManifest,
            MetadataNumberFieldManifest(
                type=FieldTypeEnum.number,
                name="fieldname",
                value="1234.1",
                unit=None,
                units=["g", "kg"],
            ),
            nullcontext(),
        ),
        (
            {
                "type": "number",
                "name": "fieldname",
                "value": "1234.1",
                "unit": "ton",
                "units": ["g", "kg"],
            },
            MetadataNumberFieldManifest,
            None,
            pytest.raises(ValueError),
        ),
        (
            {
                "type": "number",
                "name": "fieldname",
                "value": "not a number",
            },
            MetadataNumberFieldManifest,
            None,
            pytest.raises(ValueError),
        ),
        (
            {
                "type": "number",
                "name": "fieldname",
                "unit": "ton",
            },
            MetadataNumberFieldManifest,
            None,
            pytest.raises(ValueError),
        ),
        (
            {
                "type": "email",
                "name": "fieldname",
                "value": "myname@email.com",
            },
            MetadataEmailFieldManifest,
            MetadataEmailFieldManifest(
                type=FieldTypeEnum.email,
                name="fieldname",
                value="myname@email.com",
            ),
            nullcontext(),
        ),
        (
            {
                "type": "email",
                "name": "fieldname",
                "value": "notanemail",
            },
            MetadataEmailFieldManifest,
            None,
            pytest.raises(ValueError),
        ),
        (
            {
                "type": "url",
                "name": "fieldname",
                "value": "https://some.url",
            },
            MetadataURLFieldManifest,
            MetadataURLFieldManifest(
                type=FieldTypeEnum.url,
                name="fieldname",
                value="https://some.url",
            ),
            nullcontext(),
        ),
        (
            {
                "type": "url",
                "name": "fieldname",
                "value": "http://some.url",
            },
            MetadataURLFieldManifest,
            MetadataURLFieldManifest(
                type=FieldTypeEnum.url,
                name="fieldname",
                value="http://some.url",
            ),
            nullcontext(),
        ),
        (
            {
                "type": "url",
                "name": "fieldname",
                "value": "notaurl.com",
            },
            MetadataURLFieldManifest,
            None,
            pytest.raises(ValueError),
        ),
        (
            {
                "type": "date",
                "name": "fieldname",
                "value": "2024-05-28",
            },
            MetadataDateFieldManifest,
            MetadataDateFieldManifest(
                type=FieldTypeEnum.date,
                name="fieldname",
                value="2024-05-28",
            ),
            nullcontext(),
        ),
        (
            {
                "type": "date",
                "name": "fieldname",
                "value": "1234-56-78",
            },
            MetadataDateFieldManifest,
            None,
            pytest.raises(ValueError),
        ),
        (
            {
                "type": "datetime-local",
                "name": "fieldname",
                "value": "2024-05-28T15:07",
            },
            MetadataDatetimeFieldManifest,
            MetadataDatetimeFieldManifest(
                type=FieldTypeEnum.datetime_local,
                name="fieldname",
                value="2024-05-28T15:07",
            ),
            nullcontext(),
        ),
        (
            {
                "type": "datetime-local",
                "name": "fieldname",
                "value": "2024-05-28T99:99",
            },
            MetadataDatetimeFieldManifest,
            None,
            pytest.raises(ValueError),
        ),
        (
            {
                "type": "time",
                "name": "fieldname",
                "value": "15:07",
            },
            MetadataTimeFieldManifest,
            MetadataTimeFieldManifest(
                type=FieldTypeEnum.time,
                name="fieldname",
                value="15:07",
            ),
            nullcontext(),
        ),
        (
            {
                "type": "time",
                "name": "fieldname",
                "value": "99:99",
            },
            MetadataTimeFieldManifest,
            None,
            pytest.raises(ValueError),
        ),
    ),
)
def test_meta_field(
    input: dict[str, Any],
    field_type: Type[BaseMetaField],
    expected: Optional[MetadataFieldManifest],
    expectation: ContextManager,
) -> None:
    with expectation:
        result = field_type(**input)
        assert result == expected


def test_meta_field_direct_assignment() -> None:
    field = MetadataSelectFieldManifest(
        name="field",
        options=[
            "option 1",
            "option 2",
        ],
        value="option 1",
    )
    field.value = "option 2"
    assert field.value == "option 2"

    with pytest.raises(ValueError):
        field.value = "bad option"


def test_meta_field_non_existing_assignment() -> None:
    field = MetadataSelectFieldManifest(
        name="field",
        options=[
            "option 1",
            "option 2",
        ],
        value="option 1",
    )
    with pytest.raises(AttributeError):
        field.unit == "not an option"  # type: ignore


def test_group_manifest() -> None:
    data = """\
group_name: group 1
sub_fields:
- name: field 1 group 1
  type: select
  value: option 1
  options:
   - option 1
   - option 2
- name: field 2 group 1
  type: number
"""
    parsed = yaml.safe_load(data)
    verified = MetadataGroupManifest(**parsed)
    expected = MetadataGroupManifest(
        group_name="group 1",
        sub_fields=[
            MetadataSelectFieldManifest(
                name="field 1 group 1",
                value="option 1",
                type=FieldTypeEnum.select,
                options=["option 1", "option 2"],
            ),
            MetadataNumberFieldManifest(
                name="field 2 group 1",
                type=FieldTypeEnum.number,
            ),
        ],
    )
    assert verified == expected


def test_group_manifest_non_existing_subtype() -> None:
    data = """\
group_name: group 1
sub_fields:
- name: field
  type: nonexisting
"""
    parsed = yaml.safe_load(data)
    with pytest.raises(ValueError):
        MetadataGroupManifest(**parsed)


class TestExtraFieldsManifest:
    data = """\
config:
    display_main_text: false
fields:
  - name: field 1 group 1
    type: select
    value: option 1
    group: group 1
    options:
     - option 1
     - option 2
  - name: field 2 group 1
    type: number
    group: group 1
  - name: field 1 group 2
    type: text
    group: group 2
  - name: field 1 ungrouped
    type: text
"""
    parsed = yaml.safe_load(data)

    def test_extra_fields_manifest(self) -> None:
        verified = ExtraFieldsManifest(**self.parsed)
        expected = ExtraFieldsManifest(
            config=MetadataManifestConfig(
                display_main_text=False,
            ),
            fields=[
                MetadataSelectFieldManifest(
                    name="field 1 group 1",
                    value="option 1",
                    type=FieldTypeEnum.select,
                    group="group 1",
                    options=["option 1", "option 2"],
                ),
                MetadataNumberFieldManifest(
                    name="field 2 group 1",
                    type=FieldTypeEnum.number,
                    group="group 1",
                ),
                MetadataTextFieldManifest(
                    name="field 1 group 2",
                    group="group 2",
                    type=FieldTypeEnum.text,
                ),
                MetadataTextFieldManifest(
                    name="field 1 ungrouped",
                    type=FieldTypeEnum.text,
                ),
            ],
        )
        assert verified == expected

    def test_extra_fields_get_item(self) -> None:
        loaded = ExtraFieldsManifestComplex(**self.parsed).to_full_representation()
        assert loaded.field_map["field 1 group 1"] == MetadataSelectFieldManifest(
            name="field 1 group 1",
            value="option 1",
            type=FieldTypeEnum.select,
            options=["option 1", "option 2"],
            group="group 1",
        )

    def test_extra_fields_empty_manifest(self) -> None:
        verified = ExtraFieldsManifest(**{})
        expected = ExtraFieldsManifest()
        assert verified == expected

    def test_extra_fields_manifest_fail_duplicate_field(self) -> None:
        data = """\
config:
    display_main_text: false
fields:
  - name: duplicated
    type: select
    value: option 1
    group: group 1
    options:
     - option 1
     - option 2
  - name: field 2 group 1
    type: number
    group: group 1
  - name: field 1 group 2
    type: text
    group: group 2
  - name: duplicated
    type: text
"""
        parsed = yaml.safe_load(data)
        with pytest.raises(ValueError):
            ExtraFieldsManifest(**parsed).to_full_representation()

    def test_extra_fields_manifest_iter(self) -> None:
        manifest = ExtraFieldsManifest(**self.parsed).to_full_representation()
        result = [(field.group, field.name) for field in manifest.fields]
        expected = [
            ("group 1", "field 1 group 1"),
            ("group 1", "field 2 group 1"),
            ("group 2", "field 1 group 2"),
            (None, "field 1 ungrouped"),
        ]
        assert result == expected

    linked_data = """\
fields:
  - name: link 1
    type: experiments
    value: link to experiment
    group: group 1
  - name: field 2 group 1
    type: items
    group: group 1
  - name: field 1 group 2
    type: text
  - name: field 3
    type: items
    value: link to item
  - name: field 4
    type: checkbox
  - name: field 5
    type: experiments
  - name: field 6
    type: experiments
    value: link to experiment
"""
    parsed_linked = yaml.safe_load(linked_data)

    def test_extra_fields_get_dependencies(self) -> None:
        manifest = ExtraFieldsManifest(**self.parsed_linked)
        result = manifest.get_dependencies()
        expected = set(
            [
                Node(ObjectTypes.EXPERIMENT, "link to experiment"),
                Node(ObjectTypes.ITEM, "link to item"),
            ]
        )
        assert result == expected

    nested_data = """\
config:
    display_main_text: false
fields:
  - group_name: group 1
    sub_fields:
    - name: field 1 group 1
      type: select
      value: option 1
      options:
       - option 1
       - option 2
    - name: field 2 group 1
      type: number
      group: group 1
  - name: field 1 group 2
    type: text
    group: group 2
  - name: field 1 ungrouped
    type: text
"""
    parsed_nested = yaml.safe_load(nested_data)

    def test_nested_extra_fields_manifest(self) -> None:
        verified = ExtraFieldsManifestComplex(**self.parsed_nested)
        expected = ExtraFieldsManifestComplex(
            config=MetadataManifestConfig(
                display_main_text=False,
            ),
            fields=[
                MetadataGroupManifest(
                    group_name="group 1",
                    sub_fields=[
                        MetadataSelectFieldManifest(
                            name="field 1 group 1",
                            value="option 1",
                            type=FieldTypeEnum.select,
                            options=["option 1", "option 2"],
                        ),
                        MetadataNumberFieldManifest(
                            name="field 2 group 1",
                            type=FieldTypeEnum.number,
                            group="group 1",
                        ),
                    ],
                ),
                MetadataTextFieldManifest(
                    name="field 1 group 2",
                    group="group 2",
                    type=FieldTypeEnum.text,
                ),
                MetadataTextFieldManifest(
                    name="field 1 ungrouped",
                    type=FieldTypeEnum.text,
                ),
            ],
        )
        assert verified == expected

    def test_nested_extra_fields_manifest_to_full_repr_wrong_group(self) -> None:
        data = """\
fields:
  - group_name: group 1
    sub_fields:
    - name: field 1 group 1
      type: select
      value: option 1
      options:
       - option 1
       - option 2
    - name: field 2 group 1
      type: number
      group: group 2
  - name: field 1 group 2
    type: text
    group: group 2
"""
        parsed = yaml.safe_load(data)
        validated = ExtraFieldsManifestComplex(**parsed)
        with pytest.raises(ValueError):
            validated.to_full_representation()

    simple_data = """\
config:
    display_main_text: false
field_values:
    field 1: value field 1
    field 2: value field 2
    field 3:
        value: value field 3
    field 4:
        value: 4
        unit: kg
    field 6:
        value:
            - option 1
            - option 3
    field 7:
        - option 1
        - option 2
"""
    parsed_simple_data = yaml.safe_load(simple_data)

    def test_simple_extra_field_manifest(self) -> None:
        manifest = SimpleExtraFieldsManifest(**self.parsed_simple_data)
        expected = SimpleExtraFieldsManifest(
            config=MetadataManifestConfig(
                display_main_text=False,
            ),
            field_values={
                "field 1": "value field 1",
                "field 2": "value field 2",
                "field 3": _ValueAndUnit(value="value field 3"),
                "field 4": _ValueAndUnit(value="4", unit="kg"),
                "field 6": _ValueAndUnit(value=["option 1", "option 3"]),
                "field 7": ["option 1", "option 2"],
            },
        )
        assert manifest == expected

    parent_simple = ExtraFieldsManifest(
        fields=[
            MetadataTextFieldManifest(
                name="field 1",
                value="old field 1",
            ),
            MetadataExperimentsLinkFieldManifest(
                name="field 2",
                value="old field 2",
            ),
            MetadataItemsLinkFieldManifest(
                name="field 3",
                value="old field 3",
            ),
            MetadataNumberFieldManifest(
                name="field 4",
                unit="g",
                units=["kg", "g", "ton"],
            ),
            MetadataTextFieldManifest(
                name="field 5",
                value="old field 5",
            ),
            MetadataSelectFieldManifest(
                name="field 6",
                value=[],
                options=[
                    "option 1",
                    "option 2",
                    "option 3",
                ],
                allow_multi_values=True,
            ),
            MetadataSelectFieldManifest(
                name="field 7",
                value=[],
                options=[
                    "option 1",
                    "option 2",
                    "option 3",
                ],
                allow_multi_values=True,
            ),
        ]
    )

    def test_simple_extra_field_manifest_conversion(self) -> None:
        simple = SimpleExtraFieldsManifest(**self.parsed_simple_data)
        full = simple.to_full_representation(self.parent_simple)
        expected = ExtraFieldsManifest(
            config=MetadataManifestConfig(
                display_main_text=False,
            ),
            fields=[
                MetadataTextFieldManifest(
                    name="field 1",
                    value="value field 1",
                ),
                MetadataExperimentsLinkFieldManifest(
                    name="field 2",
                    value="value field 2",
                ),
                MetadataItemsLinkFieldManifest(
                    name="field 3",
                    value="value field 3",
                ),
                MetadataNumberFieldManifest(
                    name="field 4",
                    value="4",
                    unit="kg",
                    units=["kg", "g", "ton"],
                ),
                MetadataTextFieldManifest(
                    name="field 5",
                    value="old field 5",
                ),
                MetadataSelectFieldManifest(
                    name="field 6",
                    value=["option 1", "option 3"],
                    options=[
                        "option 1",
                        "option 2",
                        "option 3",
                    ],
                    allow_multi_values=True,
                ),
                MetadataSelectFieldManifest(
                    name="field 7",
                    value=["option 1", "option 2"],
                    options=[
                        "option 1",
                        "option 2",
                        "option 3",
                    ],
                    allow_multi_values=True,
                ),
            ],
        )
        assert full == expected

    def test_simple_extra_field_manifest_conversion_bad_key(self) -> None:
        simple_data = """\
config:
    display_main_text: false
field_values:
    field 1: value field 1
    field 2: value field 2
    field 3:
        value: value field 3
    field 4:
        value: 4
        unit: kg
    field 6:
        value:
            - option 1
            - option 3
    non_existing_field:
        - option 1
        - option 2
"""
        parsed_simple_data = yaml.safe_load(simple_data)
        simple = SimpleExtraFieldsManifest(**parsed_simple_data)
        with pytest.raises(ValueError):
            simple.to_full_representation(self.parent_simple)


def test_items_type_manifest() -> None:
    data = """\
version: 1
id: items type test
kind: items_type
spec:
    title: test item
    body: something
    color: "#123456"
    extra_fields:
        fields:
          - name: field 1
            type: text
          - name: field 2
            type: items
            value: link to item
"""
    parsed = yaml.safe_load(data)
    manifest = ItemsTypeManifest(**parsed)
    expected = ItemsTypeManifest(
        version=1,
        id="items type test",
        kind=ObjectTypes.ITEMS_TYPE,
        spec=ItemsTypeSpecManifest(
            title="test item",
            body="something",
            color="#123456",
            extra_fields=ExtraFieldsManifest(
                fields=[
                    MetadataTextFieldManifest(name="field 1"),
                    MetadataItemsLinkFieldManifest(
                        name="field 2",
                        value="link to item",
                    ),
                ]
            ),
        ),
    )
    assert manifest == expected


def test_items_type_manifest_nested_extra_fields() -> None:
    data = """\
version: 1
id: items type test
kind: items_type
spec:
    title: test item
    body: something
    color: "#123456"
    extra_fields:
        fields:
          - group_name: group 1
            sub_fields:
              - name: field 1
                type: text
          - name: field 2
            type: items
            value: link to item
"""
    parsed = yaml.safe_load(data)
    manifest = ItemsTypeManifest(**parsed)
    expected = ItemsTypeManifest(
        version=1,
        id="items type test",
        kind=ObjectTypes.ITEMS_TYPE,
        spec=ItemsTypeSpecManifest(
            title="test item",
            body="something",
            color="#123456",
            extra_fields=ExtraFieldsManifestComplex(
                fields=[
                    MetadataGroupManifest(
                        group_name="group 1",
                        sub_fields=[
                            MetadataTextFieldManifest(name="field 1"),
                        ],
                    ),
                    MetadataItemsLinkFieldManifest(
                        name="field 2",
                        value="link to item",
                    ),
                ]
            ),
        ),
    )
    assert manifest == expected


def test_items_type_spec_get_dependencies() -> None:
    spec = ItemsTypeSpecManifest(
        title="test",
        extra_fields=ExtraFieldsManifest(
            fields=[
                MetadataTextFieldManifest(name="field 1"),
                MetadataItemsLinkFieldManifest(
                    name="field 2",
                    value="link to item",
                ),
                MetadataItemsLinkFieldManifest(
                    name="field 3",
                ),
                MetadataExperimentsLinkFieldManifest(
                    name="field 4",
                    value="link to experiment",
                ),
                MetadataExperimentsLinkFieldManifest(
                    name="field 5",
                ),
            ]
        ),
    )
    result = spec.get_dependencies()
    expected = set(
        (
            Node(kind=ObjectTypes.ITEM, name="link to item"),
            Node(kind=ObjectTypes.EXPERIMENT, name="link to experiment"),
        )
    )
    assert result == expected


def test_experiments_template_spec_manifest() -> None:
    data = """\
version: 1
id: experiment template test
kind: experiments_template
spec:
    title: test experiment template
    tags:
        - test tag 1
        - test tag 2
    body: something
    extra_fields:
        fields:
          - name: field 1
            type: text
          - name: field 2
            type: items
            value: link to item
"""
    parsed = yaml.safe_load(data)
    manifest = ExperimentTemplateManifest(**parsed)
    expected = ExperimentTemplateManifest(
        version=1,
        id="experiment template test",
        spec=ExperimentTemplateSpecManifest(
            title="test experiment template",
            tags=["test tag 1", "test tag 2"],
            body="something",
            extra_fields=ExtraFieldsManifest(
                fields=[
                    MetadataTextFieldManifest(name="field 1"),
                    MetadataItemsLinkFieldManifest(
                        name="field 2",
                        value="link to item",
                    ),
                ]
            ),
        ),
    )
    assert manifest == expected


def test_experiment_template_get_dependencies() -> None:
    spec = ExperimentTemplateSpecManifest(
        title="test",
        extra_fields=ExtraFieldsManifest(
            fields=[
                MetadataTextFieldManifest(name="field 1"),
                MetadataItemsLinkFieldManifest(
                    name="field 2",
                    value="link to item",
                ),
                MetadataItemsLinkFieldManifest(
                    name="field 3",
                ),
                MetadataExperimentsLinkFieldManifest(
                    name="field 4",
                    value="link to experiment",
                ),
                MetadataExperimentsLinkFieldManifest(
                    name="field 5",
                ),
            ]
        ),
    )
    result = spec.get_dependencies()
    expected = set(
        (
            Node(kind=ObjectTypes.ITEM, name="link to item"),
            Node(kind=ObjectTypes.EXPERIMENT, name="link to experiment"),
        )
    )
    assert result == expected


def test_experiment_manifest_simple_metadata() -> None:
    data = """\
version: 1
id: experiment test
kind: experiment
spec:
    title: test experiment
    template: experiment template 1
    tags:
        - test tag 1
        - test tag 2
    body: something
    extra_fields:
        field_values:
            field 1: some value
            field 2: item x
"""
    parsed = yaml.safe_load(data)
    manifest = ExperimentManifest(**parsed)
    expected = ExperimentManifest(
        version=1,
        id="experiment test",
        spec=ExperimentSpecManifestSimplifiedMetadata(
            title="test experiment",
            template="experiment template 1",
            tags=["test tag 1", "test tag 2"],
            body="something",
            extra_fields=SimpleExtraFieldsManifest(
                field_values={
                    "field 1": "some value",
                    "field 2": "item x",
                }
            ),
        ),
    )
    assert manifest == expected


def test_experiment_manifest_simple_metadata_get_dependencies() -> None:
    parent_spec = ExperimentTemplateSpecManifest(
        title="test",
        extra_fields=ExtraFieldsManifest(
            fields=[
                MetadataItemsLinkFieldManifest(
                    name="field 1",
                    value="link to item from template field 1",
                ),
                MetadataItemsLinkFieldManifest(
                    name="field 2",
                ),
                MetadataExperimentsLinkFieldManifest(
                    name="field 3",
                    value="link to experiment from template field 3",
                ),
                MetadataExperimentsLinkFieldManifest(
                    name="field 4",
                ),
                MetadataTextFieldManifest(
                    name="field 5",
                ),
            ]
        ),
    )
    experiment_spec = ExperimentSpecManifestSimplifiedMetadata(
        title="test experiment",
        template="experiment template 1",
        extra_fields=SimpleExtraFieldsManifest(
            field_values={
                "field 1": "link to item from experiment field 1",
                "field 2": "link to item from experiment field 2",
                "field 4": "link to experiment from experiment field 4",
                "field 5": "some value",
            }
        ),
    )
    rendered_spec = experiment_spec.to_full_representation(parent_spec)
    result = rendered_spec.get_dependencies()
    expected = set(
        (
            Node(kind=ObjectTypes.EXPERIMENTS_TEMPLATE, name="experiment template 1"),
            Node(kind=ObjectTypes.ITEM, name="link to item from experiment field 1"),
            Node(kind=ObjectTypes.ITEM, name="link to item from experiment field 2"),
            Node(
                kind=ObjectTypes.EXPERIMENT,
                name="link to experiment from template field 3",
            ),
            Node(
                kind=ObjectTypes.EXPERIMENT,
                name="link to experiment from experiment field 4",
            ),
        )
    )
    assert result == expected


def test_experiment_manifest_nested_metadata() -> None:
    data = """\
version: 1
id: experiment test
kind: experiment
spec:
    title: test experiment
    template: experiment template 1
    tags:
        - test tag 1
        - test tag 2
    body: something
    extra_fields:
        fields:
        - group_name: group 1
          sub_fields:
            - name: field 1
              type: text
              value: some value
        - name: field 2
          type: items
          value: item x
"""
    parsed = yaml.safe_load(data)
    manifest = ExperimentManifest(**parsed)
    expected = ExperimentManifest(
        version=1,
        id="experiment test",
        spec=ExperimentSpecManifest(
            title="test experiment",
            template="experiment template 1",
            tags=["test tag 1", "test tag 2"],
            body="something",
            extra_fields=ExtraFieldsManifestComplex(
                fields=[
                    MetadataGroupManifest(
                        group_name="group 1",
                        sub_fields=[
                            MetadataTextFieldManifest(
                                name="field 1",
                                type=FieldTypeEnum.text,
                                value="some value",
                            ),
                        ],
                    ),
                    MetadataItemsLinkFieldManifest(
                        name="field 2",
                        type=FieldTypeEnum.items,
                        value="item x",
                    ),
                ]
            ),
        ),
    )
    assert manifest == expected


def test_experiment_spec_nested_metadata_get_dependencies() -> None:
    spec = ExperimentSpecManifest(
        title="test",
        extra_fields=ExtraFieldsManifest(
            fields=[
                MetadataTextFieldManifest(name="field 1"),
                MetadataItemsLinkFieldManifest(
                    name="field 2",
                    value="link to item",
                ),
                MetadataItemsLinkFieldManifest(
                    name="field 3",
                ),
                MetadataExperimentsLinkFieldManifest(
                    name="field 4",
                    value="link to experiment",
                ),
                MetadataExperimentsLinkFieldManifest(
                    name="field 5",
                ),
            ]
        ),
    )
    result = spec.get_dependencies()
    expected = set(
        (
            Node(kind=ObjectTypes.ITEM, name="link to item"),
            Node(kind=ObjectTypes.EXPERIMENT, name="link to experiment"),
        )
    )
    assert result == expected


def test_experiment_manifest_fail() -> None:
    data = """\
version: 1
id: experiment test
kind: experiment
spec:
    title: test experiment
    extra_fields:
        field_values:
            field 1: some value
            field 2: item x
"""
    parsed = yaml.safe_load(data)
    with pytest.raises(ValueError):
        ExperimentManifest(**parsed).render_spec(None)


def test_item_manifest() -> None:
    data = """\
version: 1
id: item test
kind: item
spec:
    title: test item
    category: item type 1
    tags:
        - test tag 1
        - test tag 2
    body: something
    extra_fields:
        field_values:
            field 1: some value
            field 2: item x
"""
    parsed = yaml.safe_load(data)
    manifest = ItemManifest(**parsed)
    expected = ItemManifest(
        version=1,
        id="item test",
        spec=ItemSpecManifestSimplifiedMetadata(
            title="test item",
            category="item type 1",
            tags=["test tag 1", "test tag 2"],
            body="something",
            extra_fields=SimpleExtraFieldsManifest(
                field_values={
                    "field 1": "some value",
                    "field 2": "item x",
                }
            ),
        ),
    )
    assert manifest == expected


def test_item_complex_metadata_get_dependencies() -> None:
    spec = ItemSpecManifest(
        title="test",
        category="item template",
        extra_fields=ExtraFieldsManifest(
            fields=[
                MetadataTextFieldManifest(name="field 1"),
                MetadataItemsLinkFieldManifest(
                    name="field 2",
                    value="link to item",
                ),
                MetadataItemsLinkFieldManifest(
                    name="field 3",
                ),
                MetadataExperimentsLinkFieldManifest(
                    name="field 4",
                    value="link to experiment",
                ),
                MetadataExperimentsLinkFieldManifest(
                    name="field 5",
                ),
            ]
        ),
    )
    result = spec.get_dependencies()
    expected = set(
        (
            Node(kind=ObjectTypes.ITEMS_TYPE, name="item template"),
            Node(kind=ObjectTypes.ITEM, name="link to item"),
            Node(kind=ObjectTypes.EXPERIMENT, name="link to experiment"),
        )
    )
    assert result == expected


def test_item_manifest_simple_metadata_get_dependencies() -> None:
    parent_spec = ItemsTypeSpecManifest(
        title="test",
        extra_fields=ExtraFieldsManifest(
            fields=[
                MetadataItemsLinkFieldManifest(
                    name="field 1",
                    value="link to item from template field 1",
                ),
                MetadataItemsLinkFieldManifest(
                    name="field 2",
                ),
                MetadataExperimentsLinkFieldManifest(
                    name="field 3",
                    value="link to experiment from template field 3",
                ),
                MetadataExperimentsLinkFieldManifest(
                    name="field 4",
                ),
                MetadataTextFieldManifest(
                    name="field 5",
                ),
            ]
        ),
    )
    item_spec = ItemSpecManifestSimplifiedMetadata(
        title="test item",
        category="items type 1",
        extra_fields=SimpleExtraFieldsManifest(
            field_values={
                "field 1": "link to item from experiment field 1",
                "field 2": "link to item from experiment field 2",
                "field 4": "link to experiment from experiment field 4",
                "field 5": "some value",
            }
        ),
    )
    rendered = item_spec.to_full_representation(parent_spec)
    result = rendered.get_dependencies()
    expected = set(
        (
            Node(kind=ObjectTypes.ITEMS_TYPE, name="items type 1"),
            Node(kind=ObjectTypes.ITEM, name="link to item from experiment field 1"),
            Node(kind=ObjectTypes.ITEM, name="link to item from experiment field 2"),
            Node(
                kind=ObjectTypes.EXPERIMENT,
                name="link to experiment from template field 3",
            ),
            Node(
                kind=ObjectTypes.EXPERIMENT,
                name="link to experiment from experiment field 4",
            ),
        )
    )
    assert result == expected


class TestElabObjManifests:
    data = """\
- version: 1
  id: item 1
  kind: item
  spec:
      title: test item 1
      category: item type 1
      extra_fields:
          field_values:
              field 1: some value
              field 2: item 2

- version: 1
  id: item 2
  kind: item
  spec:
      title: test item 2
      category: item type 1

- version: 1
  id: experiment 1
  kind: experiment
  spec:
      title: test experiment 1

- version: 1
  id: experiment 2
  kind: experiment
  spec:
      title: test experiment 2
      template: experiment template 1

- version: 1
  id: item type 1
  kind: items_type
  spec:
      title: test item type
      extra_fields:
          fields:
          - name: field 1
            type: text
          - name: field 2
            type: items
          - name: field 3
            type: experiments
            value: experiment 1

- version: 1
  id: experiment template 1
  kind: experiments_template
  spec:
      title: test experiments template
      extra_fields:
          fields:
          - group_name: group 1
            sub_fields:
              - name: field 4
                type: text
              - name: field 5
                type: items
                value: item 1
          - name: field 6
            type: experiments
"""
    parsed_data = yaml.safe_load(data)

    def test_elabobjmanifests(self) -> None:
        manifest = ElabObjManifests(manifests=self.parsed_data)
        expected = ElabObjManifests(
            manifests=[
                ItemManifest(
                    version=1,
                    id="item 1",
                    kind=ObjectTypes.ITEM,
                    spec=ItemSpecManifestSimplifiedMetadata(
                        title="test item 1",
                        category="item type 1",
                        extra_fields=SimpleExtraFieldsManifest(
                            field_values={
                                "field 1": "some value",
                                "field 2": "item 2",
                            }
                        ),
                    ),
                ),
                ItemManifest(
                    version=1,
                    id="item 2",
                    kind=ObjectTypes.ITEM,
                    spec=ItemSpecManifest(
                        title="test item 2",
                        category="item type 1",
                        extra_fields=None,
                    ),
                ),
                ExperimentManifest(
                    version=1,
                    id="experiment 1",
                    kind=ObjectTypes.EXPERIMENT,
                    spec=ExperimentSpecManifest(title="test experiment 1"),
                ),
                ExperimentManifest(
                    version=1,
                    id="experiment 2",
                    kind=ObjectTypes.EXPERIMENT,
                    spec=ExperimentSpecManifest(
                        title="test experiment 2",
                        template="experiment template 1",
                    ),
                ),
                ItemsTypeManifest(
                    version=1,
                    id="item type 1",
                    kind=ObjectTypes.ITEMS_TYPE,
                    spec=ItemsTypeSpecManifest(
                        title="test item type",
                        extra_fields=ExtraFieldsManifest(
                            fields=[
                                MetadataTextFieldManifest(name="field 1"),
                                MetadataItemsLinkFieldManifest(name="field 2"),
                                MetadataExperimentsLinkFieldManifest(
                                    name="field 3",
                                    value="experiment 1",
                                ),
                            ]
                        ),
                    ),
                ),
                ExperimentTemplateManifest(
                    version=1,
                    id="experiment template 1",
                    kind=ObjectTypes.EXPERIMENTS_TEMPLATE,
                    spec=ExperimentTemplateSpecManifest(
                        title="test experiments template",
                        extra_fields=ExtraFieldsManifestComplex(
                            fields=[
                                MetadataGroupManifest(
                                    group_name="group 1",
                                    sub_fields=[
                                        MetadataTextFieldManifest(name="field 4"),
                                        MetadataItemsLinkFieldManifest(
                                            name="field 5",
                                            value="item 1",
                                        ),
                                    ],
                                ),
                                MetadataExperimentsLinkFieldManifest(name="field 6"),
                            ]
                        ),
                    ),
                ),
            ]
        )
        assert manifest == expected


class TestManifestIndex:
    data = """\
- version: 1
  id: item 1
  kind: item
  spec:
      title: test item 1
      category: item type 1
      extra_fields:
          field_values:
              field 1: some value
              field 2: item 2

- version: 1
  id: item 2
  kind: item
  spec:
      title: test item 2
      category: item type 1

- version: 1
  id: experiment 1
  kind: experiment
  spec:
      title: test experiment 1

- version: 1
  id: experiment 2
  kind: experiment
  spec:
      title: test experiment 2
      template: experiment template 1

- version: 1
  id: item type 1
  kind: items_type
  spec:
      title: test item type
      extra_fields:
          fields:
          - name: field 1
            type: text
          - name: field 2
            type: items
          - name: field 3
            type: experiments
            value: experiment 1

- version: 1
  id: experiment template 1
  kind: experiments_template
  spec:
      title: test experiments template
      extra_fields:
          fields:
          - group_name: group 1
            sub_fields:
              - name: field 4
                type: text
              - name: field 5
                type: items
                value: item 1
          - name: field 6
            type: experiments
"""
    parsed_data = yaml.safe_load(data)

    def test_manifestindex_from_manifests(self) -> None:
        manifest = ElabObjManifests(manifests=self.parsed_data)
        index = ManifestIndex.from_manifests(manifest.manifests)
        expected = ManifestIndex(
            parents={
                Node(ObjectTypes.ITEM, "item 1"): Node(
                    ObjectTypes.ITEMS_TYPE, "item type 1"
                ),
                Node(ObjectTypes.ITEM, "item 2"): Node(
                    ObjectTypes.ITEMS_TYPE, "item type 1"
                ),
                Node(ObjectTypes.EXPERIMENT, "experiment 2"): Node(
                    ObjectTypes.EXPERIMENTS_TEMPLATE, "experiment template 1"
                ),
            },
            specs={
                Node(ObjectTypes.ITEM, "item 1"): ItemSpecManifest(
                    title="test item 1",
                    category="item type 1",
                    extra_fields=ExtraFieldsManifest(
                        fields=[
                            MetadataTextFieldManifest(
                                name="field 1",
                                value="some value",
                            ),
                            MetadataItemsLinkFieldManifest(
                                name="field 2",
                                value="item 2",
                            ),
                            MetadataExperimentsLinkFieldManifest(
                                name="field 3",
                                value="experiment 1",
                            ),
                        ]
                    ),
                ),
                Node(ObjectTypes.ITEM, "item 2"): ItemSpecManifest(
                    title="test item 2",
                    category="item type 1",
                    extra_fields=None,
                ),
                Node(ObjectTypes.EXPERIMENT, "experiment 1"): ExperimentSpecManifest(
                    title="test experiment 1"
                ),
                Node(ObjectTypes.EXPERIMENT, "experiment 2"): ExperimentSpecManifest(
                    title="test experiment 2",
                    template="experiment template 1",
                ),
                Node(ObjectTypes.ITEMS_TYPE, "item type 1"): ItemsTypeSpecManifest(
                    title="test item type",
                    extra_fields=ExtraFieldsManifest(
                        fields=[
                            MetadataTextFieldManifest(name="field 1"),
                            MetadataItemsLinkFieldManifest(name="field 2"),
                            MetadataExperimentsLinkFieldManifest(
                                name="field 3",
                                value="experiment 1",
                            ),
                        ]
                    ),
                ),
                Node(
                    ObjectTypes.EXPERIMENTS_TEMPLATE, "experiment template 1"
                ): ExperimentTemplateSpecManifest(
                    title="test experiments template",
                    extra_fields=ExtraFieldsManifestComplex(
                        fields=[
                            MetadataGroupManifest(
                                group_name="group 1",
                                sub_fields=[
                                    MetadataTextFieldManifest(
                                        name="field 4",
                                    ),
                                    MetadataItemsLinkFieldManifest(
                                        name="field 5",
                                        value="item 1",
                                    ),
                                ],
                            ),
                            MetadataExperimentsLinkFieldManifest(name="field 6"),
                        ]
                    ),
                ),
            },
            dependency_graph=DependencyGraph(
                edges={
                    Node(kind=ObjectTypes.ITEM, name="item 1"): set(
                        [
                            Node(kind=ObjectTypes.ITEM, name="item 2"),
                            Node(kind=ObjectTypes.ITEMS_TYPE, name="item type 1"),
                            Node(kind=ObjectTypes.EXPERIMENT, name="experiment 1"),
                        ]
                    ),
                    Node(kind=ObjectTypes.ITEM, name="item 2"): set(
                        [
                            Node(kind=ObjectTypes.ITEMS_TYPE, name="item type 1"),
                        ]
                    ),
                    Node(kind=ObjectTypes.EXPERIMENT, name="experiment 1"): set(),
                    Node(kind=ObjectTypes.EXPERIMENT, name="experiment 2"): set(
                        [
                            Node(
                                kind=ObjectTypes.EXPERIMENTS_TEMPLATE,
                                name="experiment template 1",
                            ),
                        ]
                    ),
                    Node(kind=ObjectTypes.ITEMS_TYPE, name="item type 1"): set(
                        [
                            Node(kind=ObjectTypes.EXPERIMENT, name="experiment 1"),
                        ]
                    ),
                    Node(
                        kind=ObjectTypes.EXPERIMENTS_TEMPLATE,
                        name="experiment template 1",
                    ): set(
                        [
                            Node(kind=ObjectTypes.ITEM, name="item 1"),
                        ]
                    ),
                },
                flexible=False,
            ),
        )
        assert index == expected



from __future__ import annotations

import csv
import io
import sys
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Iterable, Optional, Protocol, Sequence

import pandas as pd

from elabftwcontrol.core.interfaces import Pathlike
from elabftwcontrol.s3_utils import ParsedS3Path, write_lines_s3
from elabftwcontrol.utils import number_to_base


class OutputFormats(str, Enum):
    JSON = "json"
    CSV = "csv"
    EXCEL = "excel"
    PARQUET = "parquet"


class _LineWriterInterface(Protocol):
    def __call__(self, lines: Iterable[str]) -> None: ...


class LineWriter:
    """Callable class for writing lines to a text file.

    Instantiate using the `new` class method and pass a path.
    If no path is passed, the output will be directed to stdout.
    A path that is prefixed with s3:// will be written to AWS S3 if the
    extra `aws` dependencies are installed.
    """

    def __init__(self, line_writer: _LineWriterInterface) -> None:
        self.line_writer = line_writer

    def __call__(self, lines: Iterable[str]) -> None:
        self.line_writer(lines)

    @classmethod
    def new(cls, path: Optional[str] = None) -> LineWriter:
        writer = cls._select_writer_based_on_output(path)
        return cls(line_writer=writer)

    @classmethod
    def _select_writer_based_on_output(
        cls,
        path: Optional[str],
    ) -> _LineWriterInterface:
        writer: _LineWriterInterface
        if path is None:
            writer = _StdOutWriter()
        elif str(path).startswith("s3://"):
            writer = _S3TextWriter.new(path)
        else:
            writer = _TextFileWriter(path)
        return writer


class _StdOutWriter:
    def __call__(self, lines: Iterable[str]) -> None:
        for line in lines:
            print(line)


class _TextFileWriter:
    def __init__(
        self,
        path: Pathlike,
    ) -> None:
        self.path = path

    def __call__(self, lines: Iterable[str]) -> None:
        with open(self.path, "w") as f:
            for line in lines:
                f.write(line)
                f.write("\n")


class _S3TextWriter:
    def __init__(
        self,
        bucket: str,
        prefix: str,
    ) -> None:
        self.bucket = bucket
        self.prefix = prefix

    @classmethod
    def new(cls, path: str) -> _S3TextWriter:
        parsed_path = ParsedS3Path.from_path(path)
        return cls(
            bucket=parsed_path.bucket,
            prefix=parsed_path.prefix,
        )

    def __call__(self, lines: Iterable[str]) -> None:
        write_lines_s3(
            bucket=self.bucket,
            prefix=self.prefix,
            lines=lines,
        )


class _CSVWriterInterface(Protocol):
    def __call__(self, rows: Iterable[dict[str, Any]]) -> None: ...


class CSVWriter:
    """Callable class for rows to a CSV file.

    Instantiate using the `new` class method and pass a path and column names.
    If no path is passed, the output will be directed to stdout.
    A path that is prefixed with s3:// will be written to AWS S3 if the
    extra `aws` dependencies are installed.

    Row data is passed as dictionaries, with the keys being the column
    names and the values being the values in the row.
    """

    def __init__(
        self,
        header: Sequence[str],
        writer: _CSVWriterInterface,
    ) -> None:
        self.header = header
        self.writer = writer

    def __call__(self, rows: Iterable[dict[str, Any]]) -> None:
        self.writer(rows)

    @classmethod
    def new(
        cls,
        header: Sequence[str],
        path: Optional[Pathlike] = None,
    ) -> CSVWriter:
        writer = cls._select_writer_based_on_output(path, header=header)
        return cls(
            header=header,
            writer=writer,
        )

    @classmethod
    def _select_writer_based_on_output(
        cls,
        path: Optional[Pathlike],
        header: Sequence[str],
    ) -> _CSVWriterInterface:
        writer: _CSVWriterInterface
        if path is None:
            writer = _CSVToStdOutWriter(header=header)
        elif str(path).startswith("s3://"):
            raise NotImplementedError("Currently can't store CSVs directly on S3.")
        else:
            writer = _CSVToFileWriter(path=path, header=header)
        return writer

    @classmethod
    def write_rows(
        cls,
        file: Any,
        header: Sequence[str],
        rows: Iterable[dict[str, Any]],
    ) -> None:
        writer = csv.DictWriter(
            f=file,
            fieldnames=header,
            quoting=csv.QUOTE_NONNUMERIC,
            extrasaction="ignore",
        )
        writer.writeheader()
        for row in rows:
            writer.writerow(row)


class _CSVToStdOutWriter:
    def __init__(
        self,
        header: Sequence[str],
    ) -> None:
        self.header = header

    def __call__(
        self,
        rows: Iterable[dict[str, Any]],
    ) -> None:
        CSVWriter.write_rows(
            file=sys.stdout,
            header=self.header,
            rows=rows,
        )


class _CSVToFileWriter:
    def __init__(
        self,
        path: Pathlike,
        header: Sequence[str],
    ) -> None:
        self.path = Path(path).expanduser()
        self.header = header

    def __call__(
        self,
        rows: Iterable[dict[str, Any]],
    ) -> None:
        with open(self.path, "w", newline="") as f:
            CSVWriter.write_rows(
                file=f,
                header=self.header,
                rows=rows,
            )


PandasOutputFormats = OutputFormats


class PandasWriter:
    """Write a pandas dataframe to a CSV or Parquet file"""

    def __init__(self, writer: Callable[[pd.DataFrame], None]) -> None:
        self.writer = writer

    def __call__(self, data: pd.DataFrame) -> None:
        self.writer(data)

    @classmethod
    def new(
        cls,
        path: Optional[Pathlike] = None,
        format: PandasOutputFormats = PandasOutputFormats.CSV,
        glue_info: Optional[str] = None,
    ) -> PandasWriter:
        writer: Callable[[pd.DataFrame], None]
        if path is None:
            writer = _PandasToStdOutWriter(format)
        elif str(path).startswith("s3://"):
            writer = _PandasToS3Writer.new(
                path=str(path),
                format=format,
                glue_info=glue_info,
            )
        else:
            writer = _PandasToFileWriter(path, format)
        return cls(writer=writer)

    @classmethod
    def write(
        cls,
        data: pd.DataFrame,
        path: Pathlike | io.BytesIO,
        format: PandasOutputFormats,
    ) -> None:
        if format == OutputFormats.CSV:
            data.to_csv(
                path,
                quoting=csv.QUOTE_NONNUMERIC,
                header=True,
                index=False,
                escapechar="\\",
            )
        elif format == OutputFormats.PARQUET:
            data.to_parquet(path=path)
        else:
            raise ValueError(f"Format '{format}' not saveable as pandas dataframe.")


class _PandasToS3Writer:
    """Very basic pandas DF to S3, no partitioning"""

    def __init__(self, write_method: Callable[[pd.DataFrame], None]) -> None:
        self.write_method = write_method

    def __call__(self, df: pd.DataFrame) -> None:
        self.write_method(df)

    @classmethod
    def new(
        cls,
        path: str,
        format: PandasOutputFormats,
        glue_info: Optional[str] = None,
    ) -> _PandasToS3Writer:
        try:
            from elabftwcontrol.s3_utils import wr
        except ImportError:
            raise RuntimeError(
                "You must install optional dependencies for interacting with AWS."
            )

        common_arguments: dict[str, Any] = {
            "path": path,
            "index": False,
        }

        if glue_info is not None:
            glue_db, glue_table = glue_info.split(":")
            common_arguments.update(
                {
                    "sanitize_columns": True,
                    "dataset": True,
                    "mode": "overwrite",
                    "database": glue_db,
                    "table": glue_table,
                }
            )

        if format == OutputFormats.CSV:

            def write_method(df: pd.DataFrame) -> None:
                wr.s3.to_csv(
                    df=df,
                    **common_arguments,
                )

        elif format == OutputFormats.PARQUET:

            def write_method(df: pd.DataFrame) -> None:
                wr.s3.to_parquet(
                    df=df,
                    compression="snappy",
                    **common_arguments,
                )

        else:
            raise ValueError(f"Output format '{format}' is not supported")

        return cls(write_method)


class _PandasToFileWriter:
    def __init__(
        self,
        path: Pathlike,
        format: PandasOutputFormats,
    ) -> None:
        self.path = Path(path).expanduser()
        self.format = format

    def __call__(
        self,
        data: pd.DataFrame,
    ) -> None:
        PandasWriter.write(
            data=data,
            path=self.path,
            format=self.format,
        )


class _PandasToStdOutWriter:
    def __init__(
        self,
        format: PandasOutputFormats,
    ) -> None:
        self.format = format

    def __call__(
        self,
        data: pd.DataFrame,
    ) -> None:
        buffer = io.BytesIO()
        PandasWriter.write(
            data=data,
            path=buffer,
            format=self.format,
        )
        sys.stdout.buffer.write(buffer.getvalue())


class _SplitDataFrame(Protocol):
    key: str
    data: pd.DataFrame


class ExcelWriter:
    """Writes multiple dataframes to different sheets in one Excel file"""

    def __init__(self, writer: Callable[[Iterable[_SplitDataFrame]], None]) -> None:
        self.writer = writer

    def __call__(
        self,
        sheets: Iterable[_SplitDataFrame],
    ) -> None:
        self.writer(sheets)

    @classmethod
    def new(cls, path: Optional[Pathlike] = None) -> ExcelWriter:
        writer: Callable[[Iterable[_SplitDataFrame]], None]
        if path is None:
            writer = _ExcelToStdOutWriter()
        elif str(path).startswith("s3://"):
            raise NotImplementedError("Currently can't store Excels directly on S3.")
        else:
            writer = _ExcelToFileWriter(path=path)
        return cls(writer=writer)

    @classmethod
    def write(
        cls,
        excelwriter: pd.ExcelWriter,
        sheets: Iterable[_SplitDataFrame],
    ) -> None:
        with excelwriter:
            for sheet in sheets:
                df = sheet.data
                df.to_excel(
                    excelwriter,
                    sheet_name=sheet.key,
                    index=False,
                )
                work_sheet = excelwriter.sheets[sheet.key]
                table_range = cls.get_table_range(df)
                work_sheet.add_table(table_range, cls.get_table_options(df))

    @classmethod
    def get_table_range(cls, df: pd.DataFrame) -> str:
        """Get the data range in Excel based on the dataframe shape"""
        n_rows, n_cols = df.shape
        alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        digits = [digit - 1 for digit in number_to_base(n_cols - 1, 26)]
        digits[-1] += 1
        last_col = "".join((alphabet[i] for i in digits))
        return f"A1:{last_col}{n_rows + 1}"

    @classmethod
    def get_table_options(cls, df: pd.DataFrame) -> dict[str, Any]:
        return {"columns": [{"header": column} for column in df.columns]}


class _ExcelToStdOutWriter:
    def __call__(
        self,
        sheets: Iterable[_SplitDataFrame],
    ) -> None:
        buffer = io.BytesIO()
        writer = pd.ExcelWriter(
            buffer,
            engine="xlsxwriter",
        )
        ExcelWriter.write(writer, sheets)
        sys.stdout.buffer.write(buffer.getvalue())


class _ExcelToFileWriter:
    def __init__(
        self,
        path: Pathlike,
    ) -> None:
        self.path = Path(path).expanduser()

    def __call__(
        self,
        sheets: Iterable[_SplitDataFrame],
    ) -> None:
        writer = pd.ExcelWriter(
            self.path,
            engine="xlsxwriter",
        )
        ExcelWriter.write(writer, sheets)



import math
from typing import Any


def assert_dicts_equal(
    result: dict[str, Any],
    expected: dict[str, Any],
    order_is_important: bool = False,
) -> None:
    """Utility to check whether two dictionaries are the same even if a value of NaN"""
    if order_is_important:
        assert list(result.keys()) == list(expected.keys())
    else:
        assert set(result.keys()) == set(expected.keys())
    for field_r, value_r in result.items():
        value_e = expected[field_r]
        if isinstance(value_r, float) and math.isnan(value_r):
            assert math.isnan(value_e)
        else:
            assert value_r == value_e



from elabftwcontrol.s3_utils import ParsedS3Path


def test_parsed_s3_path() -> None:
    test_path = "s3://mybucket/my/prefix/some/file"
    path = ParsedS3Path.from_path("s3://mybucket/my/prefix/some/file")
    assert path.bucket == "mybucket"
    assert path.prefix == "my/prefix/some/file"
    assert path.path == test_path
    assert path.stripped_path == "mybucket/my/prefix/some/file"



import pandas as pd

import elabftwcontrol.utils as utils


def test_sanitize_columns() -> None:
    result = list(
        utils.sanitize_column_names(
            pd.DataFrame({"A b C": [], "g F   Hsdf": []})
        ).columns
    )
    expected = ["a_b_c", "g_f_hsdf"]
    assert result == expected


def test_join_dict_values() -> None:
    d1 = {
        "a": 1,
        "b": 1,
        "d": 1,
    }
    d2 = {
        "a": 2,
        "b": 2,
        "c": 2,
    }
    d3 = {
        "a": 3,
        "c": 3,
        "e": 3,
    }
    result = utils.join_dict_values((d1, d2, d3))
    expected = {
        "a": [1, 2, 3],
        "b": [1, 2, None],
        "c": [None, 2, 3],
        "d": [1, None, None],
        "e": [None, None, 3],
    }
    assert result == expected



from __future__ import annotations

import re
import warnings
from pathlib import Path
from typing import Any, Iterable, Optional, Sequence, TypeVar, Union

import pandas as pd
import yaml

Pathlike = Union[Path, str]

T = TypeVar("T")
V = TypeVar("V")


def number_to_base(n: int, base: int) -> list[int]:
    if n == 0:
        return [0]
    digits = []
    while n:
        digits.append(int(n % base))
        n //= base
    return digits[::-1]


def read_yaml(filepath: Pathlike) -> Union[dict, list]:
    with open(filepath, "r") as f:
        data = yaml.safe_load(f)
    return data


def compare_dicts(
    new_dict: dict[str, Any],
    old_dict: dict[str, Any],
    ignore_none_values: bool = False,
) -> dict[str, Any]:
    """See what has changed to a dict compared to an old version"""
    if ignore_none_values:
        new_values = set((k, v) for k, v in new_dict.items() if v is not None)
    else:
        new_values = set(new_dict.items())
    old_values = set(old_dict.items())
    changed_pairs = new_values - old_values
    difference = dict(changed_pairs)
    return difference


def sanitize_column_names(df: pd.DataFrame) -> pd.DataFrame:
    """Replace whitespace with underscore and put everything in lowercase"""

    def sanitize(col: str) -> str:
        return re.sub(r"\s+", "_", col.lower().strip())

    return df.rename(columns=sanitize)


def join_dict_values(dicts: Sequence[dict[Any, Any]]) -> dict[Any, list[Any]]:
    """Join multiple dictionaries on the key and merge the values in a list per key.
    If input dicts miss a key, None is inserted for that dict in the list.
    """
    iter_keys: Iterable[Any]
    all_output_keys = set().union(*(d.keys() for d in dicts))
    if all_output_keys == set(dicts[0].keys()):
        iter_keys = dicts[0].keys()
    else:
        warnings.warn(
            "Not all dicts have the same keys, resulting dict has random order."
        )
        iter_keys = all_output_keys

    output_dict: dict[Any, list[Any]] = {k: [] for k in iter_keys}

    for dct in dicts:
        for k in iter_keys:
            v = dct.get(k)
            output_dict[k].append(v)
    return output_dict


def parse_tag_str(data: Any) -> Optional[list[str]]:
    if isinstance(data, str):
        return data.split("|")
    return None


def parse_tag_id_str(data: Any) -> Optional[list[int]]:
    if isinstance(data, str):
        return [int(tag_id) for tag_id in data.split(",")]
    return None


def sanitize_name_for_glue(x: str) -> str:
    """Convert any string to one that is valid for Glue.

    The following rules apply:
    * shorter than 255 characters
    * only letters, numbers and underscores

    Whitespace and non alphanumeric characters are converted to underscores.
    Greek letters are spelled out with regular letters.
    If at the end underscores are doubled or trailing, those are also removed.
    """
    simpled = x.lower().replace(" ", "_")
    greek_letter_replace_map = {
        "α": "alpha_",
        "β": "beta_",
        "γ": "gamma_",
        "δ": "delta_",
        "ε": "epsilon_",
        "ζ": "zeta_",
        "η": "eta_",
        "θ": "theta_",
        "ι": "iota_",
        "κ": "kappa_",
        "λ": "lambda_",
        "μ": "mu_",
        "ν": "nu_",
        "ξ": "xi_",
        "ο": "omicron_",
        "π": "pi_",
        "ρ": "rho_",
        "σ": "sigma_",
        "ς": "sigma_",
        "τ": "tau_",
        "υ": "upsilon_",
        "φ": "phi_",
        "χ": "chi_",
        "ψ": "psi_",
        "ω": "omega_",
    }
    for greek, ascii in greek_letter_replace_map.items():
        simpled = simpled.replace(greek, ascii)
    simpled = re.sub(r"[^a-z0-9_]+", "_", simpled)
    simpled = re.sub(r"__+", "_", simpled).rstrip("_")
    simpled = simpled[: min(255, len(simpled))]
    return simpled


def do_nothing(x: T) -> T:
    return x



import logging
import sys

from elabftwcontrol.defaults import DEFAULT_LOG_LEVEL

logging.basicConfig(stream=sys.stderr)

logger = logging.getLogger("elabftwcontrol")
logger.setLevel(DEFAULT_LOG_LEVEL)


def set_log_level(
    level: str,
) -> None:
    logger.setLevel(level)



from pathlib import Path
from typing import Any, Iterable, Optional, Protocol, Union

Pathlike = Union[Path, str]


class HasIDAndMetadata(Protocol):
    id: int
    metadata: Optional[str]


class HasTags(Protocol):
    tags: Optional[str]


class Dictable(Protocol):
    def to_dict(self) -> dict[str, Any]: ...


class Category(Dictable, HasIDAndMetadata):
    title: str


class CategoriesApi(Protocol):
    def read(self, id: int) -> Category: ...

    def iter(self) -> Iterable[Category]: ...


class ElabApiObjInterface(HasIDAndMetadata, Dictable):
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...



from __future__ import annotations

import json
from dataclasses import dataclass
from enum import Enum
from typing import (
    Any,
    Callable,
    Final,
    Hashable,
    Iterable,
    Iterator,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    TypeVar,
)

import pandas as pd
from typing_extensions import Self

from elabftwcontrol._logging import logger
from elabftwcontrol.core.interfaces import Dictable, HasIDAndMetadata
from elabftwcontrol.core.manifests import ElabObjManifest
from elabftwcontrol.core.metadata import (
    MetadataField,
    ParsedMetadataToMetadataFieldList,
    ParsedMetadataToPandasDtype,
    ParsedMetadataToSimpleDict,
    TableCellContentType,
)
from elabftwcontrol.core.models import MetadataModel, ObjectTypes
from elabftwcontrol.core.parsers import MetadataParser
from elabftwcontrol.utils import do_nothing, sanitize_name_for_glue

DEFAULT_ITEM_SCHEMA: Final[dict[str, str]] = {
    "id": "Int64",
    "elabid": "string",
    "category": "Int64",
    "category_title": "string",
    "title": "string",
    "body": "string",
    "tags": "string",
    "fullname": "string",
    "created_at": "datetime64[ns]",
    "modified_at": "datetime64[ns]",
    "rating": "Int64",
}

DEFAULT_EXPERIMENT_SCHEMA: Final[dict[str, str]] = {
    "id": "Int64",
    "elabid": "string",
    "title": "string",
    "body": "string",
    "status_title": "string",
    "tags": "string",
    "fullname": "string",
    "created_at": "datetime64[ns]",
    "modified_at": "datetime64[ns]",
    "rating": "Int64",
}

DEFAULT_EXPERIMENTS_TEMPLATE_SCHEMA: Final[dict[str, str]] = {
    "id": "Int64",
    "title": "string",
    "body": "string",
    "tags": "string",
    "fullname": "string",
}

DEFAULT_ITEMS_TYPE_SCHEMA: Final[dict[str, str]] = {
    "id": "Int64",
    "title": "string",
    "body": "string",
}


T = TypeVar("T")


class YAMLTransformer:
    def __call__(
        self,
        objects: Iterable[Dictable],
    ) -> Iterator[str]:
        raise NotImplementedError


class DefinitionTransformer:
    def __call__(
        self,
        objects: Iterable[Dictable],
    ) -> Iterator[ElabObjManifest]:
        raise NotImplementedError


class JSONTransformer:
    def __init__(
        self,
        dict_transformer: Callable[[Iterable[Dictable]], Iterator[dict[str, Any]]],
        string_transformer: Callable[[Iterable[dict[str, Any]]], Iterator[str]],
    ) -> None:
        self._dict_transformer = dict_transformer
        self._string_transformer = string_transformer

    @classmethod
    def new(
        cls,
        indent: Optional[int] = 2,
    ) -> Self:
        return cls(
            dict_transformer=_DictTransformer(),
            string_transformer=_JSONStringTransformer(indent=indent),
        )

    def __call__(
        self,
        objects: Iterable[Dictable],
    ) -> Iterator[str]:
        dictionaries = self._dict_transformer(objects)
        return self._string_transformer(dictionaries)


class _DictTransformer:
    def __call__(self, objects: Iterable[Dictable]) -> Iterator[dict[str, Any]]:
        for obj in objects:
            yield obj.to_dict()


class _JSONStringTransformer:
    def __init__(
        self,
        indent: Optional[int],
    ) -> None:
        self.indent = indent

    def __call__(self, objects: Iterable[dict[str, Any]]) -> Iterator[str]:
        for obj in objects:
            yield json.dumps(obj, indent=self.indent)


class TableShapes(str, Enum):
    WIDE = "wide"
    LONG = "long"


@dataclass
class SplitDataFrame:
    key: str
    data: pd.DataFrame


class MultiPandasDataFrameTransformer:
    """Splits a raw pandas dataframe representation of API objects into multiple
    tables according to a splitting rule.

    Best instantiated via the `for_long_tables` or `for_wide_tables` factory
    methods.

    Parameters
    ----------
    splitter
        A callable that splits the dataframe into multiple tables.
    df_transform_getter
        A callable that returns a callable to transform each split dataframe based on
        the key provided.
    """

    def __init__(
        self,
        splitter: Callable[[pd.DataFrame], Iterable[SplitDataFrame]],
        df_transform_getter: Callable[[str], Callable[[pd.DataFrame], pd.DataFrame]],
    ) -> None:
        self.df_transform_getter = df_transform_getter
        self.splitter = splitter

    def __call__(self, raw_df: pd.DataFrame) -> Iterator[SplitDataFrame]:
        for split_df in self.split(raw_df):
            transformed = self.transform(split_df)
            yield SplitDataFrame(
                key=split_df.key,
                data=transformed,
            )

    def split(self, df: pd.DataFrame) -> Iterable[SplitDataFrame]:
        return self.splitter(df)

    def transform(self, split_df: SplitDataFrame) -> pd.DataFrame:
        return self.df_transform_getter(split_df.key)(split_df.data)

    @classmethod
    def for_raw_tables(
        cls,
        object_type: ObjectTypes,
        categories: Optional[Sequence[str]] = None,
    ) -> Self:
        def df_transform_getter(
            category: Hashable,
        ) -> Callable[[pd.DataFrame], pd.DataFrame]:
            return lambda df: df

        splitter = cls._get_splitter(object_type, categories)

        return cls(
            splitter=splitter,
            df_transform_getter=df_transform_getter,
        )

    @classmethod
    def for_long_tables(
        cls,
        object_type: ObjectTypes,
        categories: Optional[Sequence[str]] = None,
    ) -> Self:
        def df_transform_getter(
            category: Hashable,
        ) -> Callable[[pd.DataFrame], pd.DataFrame]:
            return PandasDataFrameMetadataTransformer.for_long_table(
                object_type=object_type,
            )

        splitter = cls._get_splitter(object_type, categories)

        return cls(
            splitter=splitter,
            df_transform_getter=df_transform_getter,
        )

    @classmethod
    def for_wide_tables(
        cls,
        object_type: ObjectTypes,
        categories_metadata_schema: Optional[
            Mapping[str, Optional[Mapping[str, Any]]]
        ] = None,
        cell_content: TableCellContentType = TableCellContentType.COMBINED,
        sanitize_column_names: bool = False,
    ) -> Self:
        def df_transform_getter(
            category: str,
        ) -> Callable[[pd.DataFrame], pd.DataFrame]:
            if categories_metadata_schema is None:
                metadata_schema = None
            else:
                metadata_schema = categories_metadata_schema.get(category)
            return PandasDataFrameMetadataTransformer.for_wide_table(
                object_type=object_type,
                cell_content=cell_content,
                metadata_schema=metadata_schema,
                sanitize_column_names=sanitize_column_names,
            )

        if categories_metadata_schema is None:
            categories_metadata_schema = {}
        splitter = cls._get_splitter(
            object_type,
            list(categories_metadata_schema.keys()),
        )

        return cls(
            splitter=splitter,
            df_transform_getter=df_transform_getter,
        )

    @classmethod
    def _get_splitter(
        cls,
        object_type: ObjectTypes,
        categories: Optional[Sequence[str]],
    ) -> Callable[[pd.DataFrame], Iterable[SplitDataFrame]]:
        match object_type:
            case ObjectTypes.ITEM:
                splitter = cls._get_single_column_splitter("category_title")
            case ObjectTypes.EXPERIMENT:
                if not categories:
                    raise ValueError(
                        "For experiments, category names that appear in experiment titles "
                        "must be supplied."
                    )
                logger.warning(
                    "Experiments are split up by template name, which must appear in the title"
                )
                splitter = cls._get_substring_splitter("title", categories)
            case _:
                splitter = cls._get_fake_splitter(single_category=object_type.value)
        return splitter

    @classmethod
    def _get_single_column_splitter(
        cls,
        column_name: str,
    ) -> Callable[[pd.DataFrame], Iterator[SplitDataFrame]]:
        def splitter(df: pd.DataFrame) -> Iterator[SplitDataFrame]:
            column = df[column_name]

            categories = column.unique()
            for category in categories:
                yield SplitDataFrame(
                    key=category,
                    data=df[column == category].reset_index(drop=True),
                )

        return splitter

    @classmethod
    def _get_substring_splitter(
        cls,
        column_name: str,
        categories: Iterable[str],
    ) -> Callable[[pd.DataFrame], Iterator[SplitDataFrame]]:
        def splitter(df: pd.DataFrame) -> Iterator[SplitDataFrame]:
            column = df[column_name]

            for category in categories:
                selector = column.str.contains(category)
                yield SplitDataFrame(
                    key=category,
                    data=df[selector].reset_index(drop=True),
                )

        return splitter

    @classmethod
    def _get_fake_splitter(
        cls,
        single_category: str,
    ) -> Callable[[pd.DataFrame], Iterator[SplitDataFrame]]:
        def splitter(df: pd.DataFrame) -> Iterator[SplitDataFrame]:
            yield SplitDataFrame(key=single_category, data=df)

        return splitter


class PandasDataFrameTransformer:
    """Convert API objects directly to a dataframe, each field corresponding to a column"""

    def __init__(
        self,
        dict_converter: Callable[[Iterable[Dictable]], Iterator[dict[str, Any]]],
    ) -> None:
        self.dict_converter = dict_converter

    def __call__(self, objects: Iterable[Dictable]) -> pd.DataFrame:
        dictionaries = _DictTransformer()(objects)
        raw_table = pd.DataFrame(dictionaries)
        logger.info(
            "Pandas dataframe with %s rows and %s columns created."
            % (raw_table.shape[0], raw_table.shape[1])
        )
        return raw_table

    @classmethod
    def new(cls) -> Self:
        return cls(
            dict_converter=_DictTransformer(),
        )


class PandasDataFrameMetadataTransformer:
    """Transform the metadata of a raw object dataframe"""

    def __init__(
        self,
        object_schema: Optional[Mapping[str, Any]],
        metadata_converter: Optional[Callable[[pd.Series[str]], pd.DataFrame]],
    ) -> None:
        self.object_schema = object_schema
        self.metadata_converter = metadata_converter

    def __call__(self, raw_df: pd.DataFrame) -> pd.DataFrame:
        df = raw_df.copy()
        metadata = df["metadata"] if "metadata" in df else None

        if self.object_schema is not None:
            new_df = df[[]].copy()
            for col, dtype in self.object_schema.items():
                new_df[col] = df[col].astype(dtype)
            df = new_df

        df = df.rename(columns=lambda x: f"_{x}")

        if self.metadata_converter is not None:
            assert metadata is not None
            long_metadata = self.metadata_converter(metadata)
            df = df.join(long_metadata, how="left").reset_index(drop=True)
        return df

    @classmethod
    def get_default_obj_schema(cls, obj_type: ObjectTypes) -> dict[str, str]:
        match obj_type:
            case ObjectTypes.ITEM:
                schema = DEFAULT_ITEM_SCHEMA
            case ObjectTypes.EXPERIMENT:
                schema = DEFAULT_EXPERIMENT_SCHEMA
            case ObjectTypes.EXPERIMENTS_TEMPLATE:
                schema = DEFAULT_EXPERIMENTS_TEMPLATE_SCHEMA
            case ObjectTypes.ITEMS_TYPE:
                schema = DEFAULT_ITEMS_TYPE_SCHEMA
            case _:
                raise ValueError(f"Object type '{obj_type}' not recognized")
        return schema

    @classmethod
    def for_long_table(
        cls,
        object_type: ObjectTypes,
    ) -> PandasDataFrameMetadataTransformer:
        """Create a transformer to produce long tables"""
        object_schema = cls.get_default_obj_schema(object_type)
        metadata_converter = LongMetadataTransformer.new()
        return cls(
            metadata_converter=metadata_converter,
            object_schema=object_schema,
        )

    @classmethod
    def for_wide_table(
        cls,
        object_type: ObjectTypes,
        cell_content: TableCellContentType = TableCellContentType.COMBINED,
        metadata_schema: Optional[Mapping[str, Any]] = None,
        sanitize_column_names: bool = False,
    ) -> PandasDataFrameMetadataTransformer:
        """Create a transformer to produce long tables"""
        metadata_converter = WideMetadataTransformer.new(
            cell_content=cell_content,
            metadata_schema=metadata_schema,
            sanitize_column_names=sanitize_column_names,
        )
        object_schema = cls.get_default_obj_schema(object_type)

        return cls(
            metadata_converter=metadata_converter,
            object_schema=object_schema,
        )


class WideMetadataTransformer:
    """Converts a pandas Series of eLabFTW metadata strings to wide format dataframe"""

    def __init__(
        self,
        metadata_field_parser: Callable[[str | None], MetadataModel],
        metadata_field_transformer: Callable[[MetadataModel], dict[str, Any]],
        metadata_schema: Optional[Mapping[str, Any]],
        metadata_schema_guesser: Callable[[Sequence[MetadataModel]], dict[str, Any]],
        column_name_sanitizer: Optional[Callable[[str], str]],
    ) -> None:
        self.metadata_field_parser = metadata_field_parser
        self.metadata_field_transformer = metadata_field_transformer
        self.metadata_schema = metadata_schema
        self.metadata_schema_guesser = metadata_schema_guesser
        self.column_name_sanitizer = column_name_sanitizer

    def __call__(self, metadata: pd.Series[str]) -> pd.DataFrame:
        parsed_metadata = [self.metadata_field_parser(_meta) for _meta in metadata]
        extra_fields = [
            self.metadata_field_transformer(_meta) for _meta in parsed_metadata
        ]
        df = pd.DataFrame(extra_fields).replace("", None)

        original_columns = df.columns

        if self.metadata_schema is None:
            metadata_schema = self.metadata_schema_guesser(parsed_metadata)
            metadata_schema = {col: metadata_schema[col] for col in original_columns}

        new_df = df[[]].copy()
        for col, dtype in metadata_schema.items():
            if col in df:
                new_df[col] = df[col].astype(dtype)
            else:
                new_df[col] = float("nan")

        if self.column_name_sanitizer is not None:
            new_df = new_df.rename(columns=self.column_name_sanitizer)

        return new_df

    @classmethod
    def new(
        cls,
        cell_content: TableCellContentType,
        metadata_schema: Optional[Mapping[str, Any]],
        sanitize_column_names: bool,
    ) -> WideMetadataTransformer:
        column_name_sanitizer = (
            sanitize_name_for_glue if sanitize_column_names else None
        )

        def metadata_schema_guesser(
            metadata: Sequence[MetadataModel],
        ) -> dict[str, str]:
            return cls.guess_schema(
                parsed_metadata=metadata,
                metadata_schema_transformer=ParsedMetadataToPandasDtype.new(
                    cell_content
                ),
            )

        return cls(
            metadata_field_parser=MetadataParser(),
            metadata_field_transformer=ParsedMetadataToSimpleDict.new(cell_content),
            metadata_schema=metadata_schema,
            metadata_schema_guesser=metadata_schema_guesser,
            column_name_sanitizer=column_name_sanitizer,
        )

    @classmethod
    def guess_schema(
        cls,
        parsed_metadata: Sequence[MetadataModel],
        metadata_schema_transformer: Callable[[MetadataModel], dict[str, str]],
    ) -> dict[str, Any]:
        """Return the schema of the pandas table based on column mode type"""
        dtypes = [metadata_schema_transformer(metadata) for metadata in parsed_metadata]
        dtype_df = pd.DataFrame(dtypes)

        most_common_type = {}
        for col in dtype_df.columns:
            mode = dtype_df[col].dropna().mode()
            most_common = mode.iloc[0] if not mode.empty else "string"
            most_common_type[col] = most_common
        return most_common_type


class LongMetadataTransformer:
    """Converts a pandas Series of eLabFTW metadata strings to long format dataframe"""

    def __init__(
        self,
        metadata_field_parser: Callable[[str], MetadataModel],
        metadata_field_transformer: Callable[[MetadataModel], list[MetadataField]],
    ) -> None:
        self.metadata_field_parser = metadata_field_parser
        self.metadata_field_transformer = metadata_field_transformer

    def __call__(self, metadata: pd.Series[str]) -> pd.DataFrame:
        def parse_and_transform(data: str) -> list[MetadataField]:
            parsed = self.metadata_field_parser(data)
            return self.metadata_field_transformer(parsed)

        extra_fields = metadata.apply(parse_and_transform).explode().dropna()
        df = (
            pd.DataFrame(
                extra_fields.tolist(),
                index=extra_fields.index,
            )
            .replace("", None)
            .rename(columns=lambda col: f"field_{col}")
        )

        return df

    @classmethod
    def new(cls) -> Self:
        return cls(
            metadata_field_parser=MetadataParser(),
            metadata_field_transformer=ParsedMetadataToMetadataFieldList(),
        )


class LazyWideTableUtils:
    @classmethod
    def create_header(
        self,
        object_columns: Sequence[str],
        metadata_columns: Sequence[str],
        column_name_sanitizer: Optional[Callable[[str], str]] = None,
    ) -> list[str]:
        header = [f"_{col}" for col in object_columns]
        if column_name_sanitizer:
            field_columns: Iterable[str] = map(
                column_name_sanitizer,
                metadata_columns,
            )
        else:
            field_columns = metadata_columns
        header.extend(field_columns)
        return header

    @classmethod
    def sanitize_cell_value(cls, cell_value: Any) -> Any:
        if isinstance(cell_value, str):
            transformed = cell_value.replace("\n", "\\n").replace('"', "")
            return transformed
        else:
            return cell_value

    @classmethod
    def transform_object_to_row_wide(
        cls,
        obj: HasIDAndMetadata,
        object_columns: Sequence[str],
        metadata_columns: Sequence[str],
        metadata_parser: Callable[[Optional[str]], MetadataModel],
        dict_converter: Callable[[MetadataModel], dict[str, Any]],
        column_name_sanitizer: Optional[Callable[[str], str]],
    ) -> dict[str, Any]:
        row_data = {}

        if column_name_sanitizer is None:
            column_name_sanitizer = do_nothing

        for column in object_columns:
            row_data[f"_{column}"] = getattr(obj, column, None)

        parsed_metadata = metadata_parser(obj.metadata)
        field_dictionary = dict_converter(parsed_metadata)

        for column in metadata_columns:
            cell_value = cls.sanitize_cell_value(field_dictionary.get(column))
            row_data[column_name_sanitizer(column)] = cell_value

        return row_data


class CSVTransformer:
    """Converts API objects data to a wide table format row by row"""

    def __init__(
        self,
        object_columns: Sequence[str],
        metadata_columns: Sequence[str],
        metadata_parser: Callable[[Optional[str]], MetadataModel],
        dict_converter: Callable[[MetadataModel], dict[str, Any]],
        column_name_sanitizer: Optional[Callable[[str], str]],
    ) -> None:
        self.object_columns = object_columns
        self.metadata_columns = metadata_columns
        self.dict_converter = dict_converter
        self.column_name_sanitizer = column_name_sanitizer
        self.metadata_parser = metadata_parser

    def __call__(
        self,
        objects: Iterable[HasIDAndMetadata],
    ) -> Iterable[dict[str, Any]]:
        return map(self._transform, objects)

    def get_header(self) -> list[str]:
        return LazyWideTableUtils.create_header(
            object_columns=self.object_columns,
            metadata_columns=self.metadata_columns,
            column_name_sanitizer=self.column_name_sanitizer,
        )

    @classmethod
    def new(
        cls,
        object_columns: Sequence[str],
        metadata_columns: Sequence[str],
        cell_content: TableCellContentType,
        sanitize_columns: bool,
    ) -> CSVTransformer:
        return cls(
            object_columns=object_columns,
            metadata_columns=metadata_columns,
            metadata_parser=MetadataParser(),
            dict_converter=ParsedMetadataToSimpleDict.new(cell_content),
            column_name_sanitizer=sanitize_name_for_glue if sanitize_columns else None,
        )

    def _transform(
        self,
        obj: HasIDAndMetadata,
    ) -> dict[str, Any]:
        return LazyWideTableUtils.transform_object_to_row_wide(
            obj=obj,
            object_columns=self.object_columns,
            metadata_columns=self.metadata_columns,
            metadata_parser=self.metadata_parser,
            dict_converter=self.dict_converter,
            column_name_sanitizer=self.column_name_sanitizer,
        )


class CSVLongMetadataTableTransformer:
    """Convert all metadata from objects to a long fixed schema table"""

    SCHEMA: tuple[str, ...] = MetadataField._fields

    def __init__(
        self,
        metadata_parser: Callable[[Optional[str]], MetadataModel],
        metadatafield_transformer: Callable[[MetadataModel], list[MetadataField]],
        field_sanitizer: Callable[[Any], Any],
    ) -> None:
        self.metadata_parser = metadata_parser
        self.metadatafield_transformer = metadatafield_transformer
        self.field_sanitizer = field_sanitizer

    @classmethod
    def new(cls) -> Self:
        return cls(
            metadata_parser=MetadataParser(),
            metadatafield_transformer=ParsedMetadataToMetadataFieldList(),
            field_sanitizer=LazyWideTableUtils.sanitize_cell_value,
        )

    def __call__(
        self,
        objects: Iterable[HasIDAndMetadata],
    ) -> Iterator[dict[str, Any]]:
        for obj in objects:
            fields = self._transform(obj)
            for field in fields:
                yield field

    def get_header(self) -> tuple[str, ...]:
        return self.SCHEMA

    def _transform(self, obj: HasIDAndMetadata) -> Iterator[dict[str, Any]]:
        """Convert a single API object to a bunch of metadata fields"""
        fields = self.metadatafield_transformer(self.metadata_parser(obj.metadata))
        for field in fields:
            field_dict = {f"field_{k}": v for k, v in field._asdict().items()}
            field_dict["_id"] = obj.id
            field_dict["field_value"] = self.field_sanitizer(field.value)
            yield field_dict


class CSVElabDataTableTransformer:
    """Convert elabftw data of objects to a table with fixed schema"""

    def __init__(
        self,
        columns: Sequence[str],
        column_name_sanitizer: Optional[Callable[[str], str]],
    ) -> None:
        self.columns = columns
        self.column_name_sanitizer = column_name_sanitizer

    def __call__(
        self,
        objects: Iterable[Any],
    ) -> Iterable[dict[str, Any]]:
        return map(self.transform, objects)

    def get_header(self) -> list[str]:
        if self.column_name_sanitizer:
            return list(map(self.column_name_sanitizer, self.columns))
        else:
            return list(self.columns)

    @classmethod
    def new(
        cls,
        columns: Sequence[str],
        sanitize_columns: bool,
    ) -> CSVElabDataTableTransformer:
        return cls(
            columns=columns,
            column_name_sanitizer=sanitize_name_for_glue if sanitize_columns else None,
        )

    def transform(
        self,
        obj: Any,
    ) -> dict[str, Any]:
        row_data = {}

        for column in self.columns:
            if hasattr(obj, column):
                renamed_column = (
                    self.column_name_sanitizer(column)
                    if self.column_name_sanitizer
                    else column
                )
                row_data[renamed_column] = getattr(obj, column)

        return row_data


class WideObjectTableData(NamedTuple):
    table_name: str
    object_columns: Sequence[str]
    metadata_columns: Sequence[str]
    objects: Iterable[HasIDAndMetadata]


class ExcelTransformer:
    def __init__(
        self,
        metadata_parser: Callable[[Optional[str]], MetadataModel],
        dict_converter: Callable[[MetadataModel], dict[str, Any]],
        column_name_sanitizer: Optional[Callable[[str], str]],
    ) -> None:
        self.metadata_parser = metadata_parser
        self.dict_converter = dict_converter
        self.column_name_sanitizer = column_name_sanitizer

    @classmethod
    def new(
        cls,
        cell_content: TableCellContentType,
        sanitize_columns: bool,
    ) -> ExcelTransformer:
        return cls(
            metadata_parser=MetadataParser(),
            dict_converter=ParsedMetadataToSimpleDict.new(cell_content),
            column_name_sanitizer=sanitize_name_for_glue if sanitize_columns else None,
        )

    def __call__(
        self,
        sheets: Iterable[WideObjectTableData],
    ) -> Iterator[SplitDataFrame]:
        return map(self.transform_sheet_to_dataframe, sheets)

    def transform_sheet_to_dataframe(
        self,
        sheet: WideObjectTableData,
    ) -> SplitDataFrame:
        header = LazyWideTableUtils.create_header(
            object_columns=sheet.object_columns,
            metadata_columns=sheet.metadata_columns,
            column_name_sanitizer=self.column_name_sanitizer,
        )
        df = pd.DataFrame(
            (
                self.transform_object_to_row(
                    obj=obj,
                    metadata_columns=sheet.metadata_columns,
                    object_columns=sheet.object_columns,
                )
                for obj in sheet.objects
            ),
            columns=header,
        )
        return SplitDataFrame(
            key=sheet.table_name,
            data=df,
        )

    def transform_object_to_row(
        self,
        obj: HasIDAndMetadata,
        object_columns: Sequence[str],
        metadata_columns: Sequence[str],
    ) -> dict[str, Any]:
        return LazyWideTableUtils.transform_object_to_row_wide(
            obj=obj,
            object_columns=object_columns,
            metadata_columns=metadata_columns,
            metadata_parser=self.metadata_parser,
            dict_converter=self.dict_converter,
            column_name_sanitizer=self.column_name_sanitizer,
        )



from __future__ import annotations

from enum import Enum
from typing import Any, Callable, NamedTuple, Optional, TypeVar

from elabftwcontrol._logging import logger
from elabftwcontrol.core.models import FieldTypeEnum, MetadataModel, SingleFieldModel

T = TypeVar("T")
V = TypeVar("V")


class TableCellContentType(str, Enum):
    """When downloading as a table, what should be in the cells"""

    VALUE = "value"
    UNIT = "unit"
    COMBINED = "combined"


class MetadataField(NamedTuple):
    """Named tuple representation for a field for easy conversion to pandas table"""

    name: str = ""
    type: str = ""
    group: str = ""
    description: str = ""
    position: int = -1
    value: str = ""
    unit: str = ""

    @classmethod
    def from_parsed_field(
        cls,
        data: SingleFieldModel,
        name: Optional[str] = None,
        group: Optional[str] = None,
    ) -> MetadataField:
        field_position = cls.safe_get_field_position(data.position)
        return cls(
            name=str(name or ""),
            type=data.type.value,
            group=str(group or ""),
            description=str(data.description or ""),
            position=field_position,
            value=str(data.value or ""),
            unit=str(data.unit or ""),
        )

    @classmethod
    def safe_get_field_position(cls, position: Any) -> int:
        try:
            return int(position)
        except Exception:
            return -1


class ParsedMetadataToMetadataFieldList:
    """Convert parsed object metadata to a list of named tuples"""

    def __call__(self, parsed_metadata: MetadataModel) -> list[MetadataField]:
        return self.transform_api_object_metadata(parsed_metadata)

    @classmethod
    def transform_api_object_metadata(
        cls,
        parsed_metadata: MetadataModel,
    ) -> list[MetadataField]:
        groups = parsed_metadata.elabftw.extra_fields_groups
        group_map = {group.id: group.name for group in groups}

        fields: list[MetadataField] = []
        for fieldname, field in parsed_metadata.extra_fields.items():
            group_id = field.group_id
            if group_id is None:
                field_group = None
            else:
                field_group = group_map.get(group_id)
            if field_group is None:
                logger.debug(f"Field {fieldname} does not have a valid group.")
                field_group = None

            new_field = MetadataField.from_parsed_field(
                data=field,
                name=fieldname,
                group=field_group,
            )
            fields.append(new_field)

        return fields


class ParsedMetadataToSimpleDict:
    """Convert parsed object metadata to a simplified key: value dictionary"""

    def __init__(
        self,
        parser: Callable[[MetadataModel], dict[str, Any]],
    ) -> None:
        self.parser = parser

    def __call__(self, parsed_metadata: MetadataModel) -> dict[str, Any]:
        return self.parser(parsed_metadata)

    @classmethod
    def new(
        cls,
        cell_content: TableCellContentType,
    ) -> ParsedMetadataToSimpleDict:
        match cell_content:
            case TableCellContentType.COMBINED:

                def parser(parsed_metadata: MetadataModel) -> dict[str, Any]:
                    return parsed_metadata.field_values_and_units

            case TableCellContentType.UNIT:

                def parser(parsed_metadata: MetadataModel) -> dict[str, Any]:
                    return parsed_metadata.field_units

            case TableCellContentType.VALUE:

                def parser(parsed_metadata: MetadataModel) -> dict[str, Any]:
                    return parsed_metadata.field_values

            case _:
                raise ValueError(f"Cell content '{cell_content}' not recognized.")

        return cls(parser=parser)


class ParsedMetadataToPandasDtype:
    def __init__(self, converter: Callable[[SingleFieldModel], str]) -> None:
        self.converter = converter

    @classmethod
    def new(cls, cell_content: TableCellContentType) -> ParsedMetadataToPandasDtype:
        match cell_content:
            case TableCellContentType.VALUE:
                converter = cls.get_pandas_value_dtype
            case TableCellContentType.COMBINED:
                converter = cls.get_pandas_combined_dtype
            case TableCellContentType.UNIT:
                converter = cls.get_pandas_unit_dtype
            case _:
                raise ValueError(f"Cell content type '{cell_content}' not recognized.")
        return cls(converter)

    def __call__(self, parsed_metadata: MetadataModel) -> dict[str, str]:
        return self.transform_api_object_metadata(
            parsed_metadata,
            converter=self.converter,
        )

    @classmethod
    def transform_api_object_metadata(
        cls,
        parsed_metadata: MetadataModel,
        converter: Callable[[SingleFieldModel], str],
    ) -> dict[str, str]:
        return parsed_metadata.map_fields(converter)

    @classmethod
    def get_pandas_unit_dtype(cls, field: SingleFieldModel) -> str:
        return "string"

    @classmethod
    def get_pandas_value_dtype(cls, field: SingleFieldModel) -> str:
        match field.type:
            case FieldTypeEnum.number:
                return "Float64"

            case FieldTypeEnum.date:
                return "datetime64[ns]"

            case FieldTypeEnum.datetime_local:
                return "datetime64[ns]"

            case FieldTypeEnum.time:
                return "datetime64[ns]"

            case FieldTypeEnum.items:
                return "Int64"

            case FieldTypeEnum.experiments:
                return "Int64"

            case _:
                return "string"

    @classmethod
    def get_pandas_combined_dtype(cls, field: SingleFieldModel) -> str:
        match field.type:
            case FieldTypeEnum.number:
                return "string"

            case FieldTypeEnum.date:
                return "datetime64[ns]"

            case FieldTypeEnum.datetime_local:
                return "datetime64[ns]"

            case FieldTypeEnum.time:
                return "datetime64[ns]"

            case FieldTypeEnum.items:
                return "Int64"

            case FieldTypeEnum.experiments:
                return "Int64"

            case _:
                return "string"



from typing import Any

from elabftwcontrol.core.models import (
    ConfigMetadata,
    FieldTypeEnum,
    GroupModel,
    MetadataModel,
    SingleFieldModel,
)
from elabftwcontrol.core.parsers import MetadataParser, TagIdParser, TagParser


class TestMetadataParser:
    def test_metadata_parser(self) -> None:
        dummy_metadata = """\
{
    "elabftw": {
        "extra_fields_groups": [
            {"id": 1, "name": "g1"},
            {"id": 2, "name": "g2"},
            {"id": 3, "name": "g3"}
        ]
    },
    "extra_fields": {
        "f1": {
            "type": "number",
            "value": "2.3",
            "unit": "kg",
            "group_id": "1",
            "position": "3"
        },
        "k3": {
            "type": "date",
            "value": "2023-09-12",
            "group_id": "1",
            "position": "2"
        }
    }
}
"""
        parser = MetadataParser()
        model = parser(dummy_metadata)
        expected = MetadataModel(
            elabftw=ConfigMetadata(
                extra_fields_groups=[
                    GroupModel(id=1, name="g1"),
                    GroupModel(id=2, name="g2"),
                    GroupModel(id=3, name="g3"),
                ]
            ),
            extra_fields={
                "f1": SingleFieldModel(
                    type=FieldTypeEnum.number,
                    value="2.3",
                    unit="kg",
                    group_id=1,
                    position=3,
                ),
                "k3": SingleFieldModel(
                    type=FieldTypeEnum.date,
                    value="2023-09-12",
                    group_id=1,
                    position=2,
                ),
            },
        )
        assert model == expected

    def test_empty_metadata(self) -> None:
        parser = MetadataParser()
        model = parser(None)
        expected = MetadataModel()
        assert model == expected

    def test_json_error_metadata(self) -> None:
        json_with_errors = "some non json metadata"
        parser = MetadataParser()
        model = parser(json_with_errors)
        expected = MetadataModel()
        assert model == expected


class TestTagParser:
    def test_normal_tags(self) -> None:
        tags = "tag1|tag2|tag3"
        parser = TagParser()
        result = parser(tags)
        expected = ["tag1", "tag2", "tag3"]
        assert result == expected

    def test_none(self) -> None:
        tags = None
        parser = TagParser()
        result = parser(tags)
        assert result == []

    def test_incorrect(self) -> None:
        tags: Any = object()
        parser = TagParser()
        result = parser(tags)
        assert result == []


class TestTagIdParser:
    def test_normal_tags(self) -> None:
        tags = "123,234,2"
        parser = TagIdParser()
        result = parser(tags)
        expected = [123, 234, 2]
        assert result == expected

    def test_none(self) -> None:
        tags = None
        parser = TagIdParser()
        result = parser(tags)
        assert result == []

    def test_incorrect(self) -> None:
        tags = "not_int,1, 2,3"
        parser = TagIdParser()
        result = parser(tags)
        assert result == []



from io import StringIO
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import Mock, patch

import pandas as pd
import pytest

from elabftwcontrol.download.output import CSVWriter, ExcelWriter, LineWriter


def test_text_to_file_writer() -> None:
    with TemporaryDirectory() as tmpdir:
        data = [
            "line1",
            "line2\nline3",
            """\
line4
line5""",
        ]
        filepath = Path(tmpdir) / "test_file.csv"

        writer = LineWriter.new(filepath)
        writer(data)

        with open(filepath, "r") as f:
            content = f.read()

        expected = """\
line1
line2
line3
line4
line5
"""
        assert expected == content


def test_text_to_stdout_writer() -> None:
    with patch("sys.stdout", new=StringIO()) as fake_out:
        data = [
            "line1",
            "line2\nline3",
            """\
line4
line5""",
        ]

        writer = LineWriter.new()
        writer(data)

        expected = """\
line1
line2
line3
line4
line5
"""
        assert expected == fake_out.getvalue()


def test_csv_writer_write_rows() -> None:
    with TemporaryDirectory() as tmpdir:
        header = ["col1", "col2"]
        data = [
            {"col1": 1, "col2": 2},
            {"col1": 4, "col2": "test"},
            {"col2": 4, "col3": 5},
        ]
        filepath = Path(tmpdir) / "test_file.csv"

        with open(filepath, "w") as f:
            CSVWriter.write_rows(file=f, header=header, rows=data)

        with open(filepath, "r") as f:
            content = f.read()

        expected = """\
\"col1\",\"col2\"
1,2
4,"test"
\"\",4
"""
        assert expected == content


def test_csv_to_file_writer() -> None:
    with TemporaryDirectory() as tmpdir:
        data = [
            {"col1": 1},
            {"col1": 1, "col2": 2},
            {"col1": 4, "col2": "test"},
            {"col2": 4, "col3": 5},
        ]
        filepath = Path(tmpdir) / "test_file.csv"

        writer = CSVWriter.new(
            path=filepath,
            header=["col1", "col2"],
        )
        writer(data)

        with open(filepath, "r") as f:
            content = f.read()

        expected = """\
\"col1\",\"col2\"
1,\"\"
1,2
4,"test"
\"\",4
"""
        assert expected == content


def test_csv_to_stdout_writer() -> None:
    with patch("sys.stdout", new=StringIO()) as fake_out:
        data = [
            {"col1": 1, "col2": 2},
            {"col1": 4, "col2": "test"},
            {"col2": 4, "col3": 5},
        ]

        writer = CSVWriter.new(header=["col1", "col2"])
        writer(data)

        expected = """\
\"col1\",\"col2\"\r
1,2\r
4,"test"\r
\"\",4\r
"""
        assert expected == fake_out.getvalue()


class TestExcelWriter:
    def test_write(self) -> None:
        pass

    @pytest.mark.parametrize(
        ("shape", "expected"),
        (
            ((0, 5), "A1:E1"),
            ((4, 26), "A1:Z5"),
            ((4, 27), "A1:AA5"),
            ((4, 28), "A1:AB5"),
            ((4, 703), "A1:AAA5"),
            ((4, 3487), "A1:EDC5"),
            ((4, 16377), "A1:XEW5"),
        ),
    )
    def test_get_table_range(
        self,
        shape: tuple[int, int],
        expected: str,
    ) -> None:
        mockdf = Mock()
        mockdf.shape = shape
        assert ExcelWriter.get_table_range(mockdf) == expected

    def test_write_file(self) -> None:
        with TemporaryDirectory() as tmpdir:
            df_1 = pd.DataFrame(
                {
                    "a": [1, 2],
                    "b": ["x", "y"],
                }
            )
            df_2 = pd.DataFrame(
                {
                    "a": [1, 2],
                    "b": ["x", "y"],
                }
            )

            output = [
                Mock(
                    key="sheet_1",
                    data=df_1,
                ),
                Mock(
                    key="sheet_2",
                    data=df_2,
                ),
            ]

            test_file = Path(tmpdir) / "test_file.xlsx"
            writer = ExcelWriter.new(test_file)
            writer(output)

            for sheet, df in zip(("sheet_1", "sheet_2"), (df_1, df_2)):
                test = pd.read_excel(test_file, sheet_name=sheet)
                pd.testing.assert_frame_equal(test, df)



from __future__ import annotations

from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Mapping, NamedTuple, Optional, Type, TypeVar, Union

from elabapi_python import (  # type: ignore
    Experiment,
    ExperimentTemplate,
    Item,
    ItemsType,
)
from pydantic import BaseModel, Field

from elabftwcontrol._logging import logger
from elabftwcontrol.core.interfaces import ElabApiObjInterface

Pathlike = Union[str, Path]


T = TypeVar("T")
V = TypeVar("V")


class IdNode(NamedTuple):
    """Minimum piece of information to identify an object"""

    kind: ObjectTypes
    id: int

    def __str__(self) -> str:
        return f"{self.kind}: {self.id}"


class NameNode(NamedTuple):
    """Minimum piece of information to identify a definition"""

    kind: ObjectTypes
    name: str

    def __str__(self) -> str:
        return f"{self.kind}: {self.name}"


class Auth(BaseModel):
    base: int = 30
    teams: list[int] = Field(default_factory=list)
    users: list[int] = Field(default_factory=list)
    teamgroups: list[int] = Field(default_factory=list)


class ObjectTypes(str, Enum):
    ITEM = "item"
    ITEMS_TYPE = "items_type"
    EXPERIMENT = "experiment"
    EXPERIMENTS_TEMPLATE = "experiments_template"

    def is_template(self) -> bool:
        return self == self.ITEMS_TYPE or self == self.EXPERIMENTS_TEMPLATE

    def is_individual(self) -> bool:
        return self == self.ITEM or self == self.EXPERIMENT


OBJTYPE_TO_CLASS: Mapping[ObjectTypes, Type[ElabApiObjInterface]] = {
    ObjectTypes.ITEM: Item,
    ObjectTypes.EXPERIMENT: Experiment,
    ObjectTypes.ITEMS_TYPE: ItemsType,
    ObjectTypes.EXPERIMENTS_TEMPLATE: ExperimentTemplate,
}

CLASS_TO_OBJTYPE: Mapping[Type[ElabApiObjInterface], ObjectTypes] = {
    Item: ObjectTypes.ITEM,
    Experiment: ObjectTypes.EXPERIMENT,
    ItemsType: ObjectTypes.ITEMS_TYPE,
    ExperimentTemplate: ObjectTypes.EXPERIMENTS_TEMPLATE,
}


class GroupModel(BaseModel):
    id: int
    name: str


class ConfigMetadata(BaseModel):
    display_main_text: Optional[bool] = None
    extra_fields_groups: list[GroupModel] = []


class FieldTypeEnum(str, Enum):
    checkbox = "checkbox"
    radio = "radio"
    select = "select"
    text = "text"
    number = "number"
    email = "email"
    url = "url"
    date = "date"
    datetime_local = "datetime-local"
    time = "time"
    items = "items"
    experiments = "experiments"


def _safe_parse(
    return_type: str,
    fail_value: V,
) -> Callable[[Callable[[Any, str], T]], Callable[[Any, str], T | V]]:
    """Decorator to ensure all parsing failures are handled"""

    def decorator(parser: Callable[[Any, str], T]) -> Callable[[Any, str], T | V]:
        def wrapper(cls: Any, value: str) -> Union[T, V]:
            try:
                return parser(cls, value)
            except Exception:
                logger.debug(f"Could not parse {value} as a {return_type}")
                return fail_value

        return wrapper

    return decorator


class SingleFieldModel(BaseModel):
    value: Optional[int | str | list[str]] = None
    type: FieldTypeEnum = FieldTypeEnum.text
    options: Optional[list[str]] = None
    allow_multi_values: Optional[bool] = None
    required: Optional[bool] = None
    description: Optional[str] = None
    units: Optional[list[str]] = None
    unit: Optional[str] = None
    position: Optional[int] = None
    blank_value_on_duplicate: Optional[bool] = None
    group_id: Optional[int] = None
    readonly: Optional[bool] = None

    @property
    def parsed_value(self) -> Any:
        return self._parse_field_value(self.value, self.type)

    @property
    def value_and_unit(self) -> str:
        """Value and unit concatenated"""
        if not self.value:
            return ""
        elif not self.unit:
            return str(self.value)
        else:
            return f"{self.value} {self.unit}"

    @property
    def corrected_unit(self) -> str:
        """If there is no unit, empty string"""
        if not self.value or not self.unit:
            return ""
        else:
            return self.unit

    @classmethod
    def _parse_field_value(cls, value: Any, field_type: FieldTypeEnum) -> Any:
        """Parse the value from a field based on the field type"""
        match field_type:
            case FieldTypeEnum.number:
                return cls.parse_number(value)

            case FieldTypeEnum.date:
                return cls.parse_date(value)

            case FieldTypeEnum.datetime_local:
                return cls.parse_datetime(value)

            case FieldTypeEnum.time:
                return cls.parse_time(value)

            case FieldTypeEnum.items:
                return cls.parse_item_link(value)

            case FieldTypeEnum.experiments:
                return cls.parse_experiment_link(value)

            case _:
                return value

    @classmethod
    @_safe_parse(return_type="number", fail_value=float("nan"))
    def parse_number(cls, value: str) -> float:
        return float(value)

    @classmethod
    @_safe_parse(return_type="date", fail_value=None)
    def parse_date(cls, value: str) -> Optional[datetime]:
        parsed = datetime.strptime(value, "%Y-%m-%d")
        return parsed

    @classmethod
    @_safe_parse(return_type="datetime", fail_value=None)
    def parse_datetime(cls, value: str) -> Optional[datetime]:
        return datetime.strptime(value, "%Y-%m-%dT%H:%M")

    @classmethod
    @_safe_parse(return_type="time", fail_value=None)
    def parse_time(cls, value: str) -> Optional[datetime]:
        return datetime.strptime(value, "%H:%M")

    @classmethod
    @_safe_parse(return_type="link to an item", fail_value=None)
    def parse_item_link(cls, value: str | int) -> Optional[int]:
        return cls.parse_link(value)

    @classmethod
    @_safe_parse(return_type="link to an experiment", fail_value=None)
    def parse_experiment_link(cls, value: str | int) -> Optional[int]:
        return cls.parse_link(value)

    @classmethod
    def parse_link(cls, value: str | int) -> int:
        return int(value)


class ElabftwControlConfig(BaseModel):
    template_name: str | None = None
    name: str | None = None
    version: str | None = None


class MetadataModel(BaseModel):
    elabftw: ConfigMetadata = ConfigMetadata()
    extra_fields: dict[str, SingleFieldModel] = {}
    elabftwcontrol: ElabftwControlConfig | None = None

    @property
    def field_dict(self) -> dict[str, dict[str, Any]]:
        """Extra fields represented as a simple nested dictionary"""
        dict_dump = self.model_dump(exclude_none=True)
        return dict_dump["extra_fields"]

    @property
    def ordered_fieldnames(self) -> list[str]:
        """Get the names of fields in order of position"""
        fields: list[tuple[int, str]] = []
        for fieldname, field in self.extra_fields.items():
            if field.position is None:
                position = -1
            else:
                try:
                    position = int(field.position)
                except ValueError:
                    logger.debug("Invalid position defined on field %s" % fieldname)
                    position = -1

            fields.append((position, fieldname))

        fields.sort()

        return [field[1] for field in fields]

    @property
    def fieldnames(self) -> list[str]:
        """Names of fields in arbitrary order"""
        return list(self.extra_fields.keys())

    def map_fields(self, accessor: Callable[[SingleFieldModel], T]) -> dict[str, T]:
        """Apply a function to all fields and return a dictionary with the results"""
        return {
            fieldname: accessor(self.extra_fields[fieldname])
            for fieldname in self.ordered_fieldnames
        }

    @property
    def field_values(self) -> dict[str, Any]:
        """Name: values of fields ordered by field position"""

        def accessor(field: SingleFieldModel) -> Any:
            return field.parsed_value

        return self.map_fields(accessor)

    @property
    def field_units(self) -> dict[str, str]:
        """Name: unit of fields ordered by field position"""

        def accessor(field: SingleFieldModel) -> str:
            return field.corrected_unit

        return self.map_fields(accessor)

    @property
    def field_values_and_units(self) -> dict[str, str]:
        """Name: value + unit as strings of fields ordered by field position"""

        def accessor(field: SingleFieldModel) -> str:
            return field.value_and_unit

        return self.map_fields(accessor)



from __future__ import annotations

from typing import Any, Dict, Generic, List, Optional, Sequence, Set, TypeVar

T = TypeVar("T")


class Edge(Generic[T]):
    source: T
    destination: T


class DependencyGraph(Generic[T]):
    def __init__(
        self,
        edges: Optional[Dict[T, Set[T]]] = None,
        flexible: bool = True,
    ) -> None:
        if edges is None:
            edges = {}
        self.graph: Dict[T, Set[T]] = edges
        if flexible:
            self.add_node = self._flexible_add_node
            self.add_edge = self._flexible_add_edge
            self.get_dependencies = self._flexible_get_dependencies
        else:
            self.add_node = self._rigid_add_node
            self.add_edge = self._rigid_add_edge
            self.get_dependencies = self._rigid_get_dependencies

    def __eq__(self, o: Any) -> bool:
        return self.graph == o.graph

    def __str__(self) -> str:
        to_return = ""
        for node, connected_nodes in self.graph.items():
            to_return += f"-{node}:"
            if not connected_nodes:
                to_return += " /\n"
            else:
                to_return += "\n"
                for connected_node in connected_nodes:
                    to_return += f"   -> {connected_node}\n"

        return to_return

    def has_node(self, node: T) -> bool:
        return node in self.graph

    def _flexible_get_dependencies(self, node: T) -> Set[T]:
        return self.graph.get(node, set())

    def _rigid_get_dependencies(self, node: T) -> Set[T]:
        return self.graph[node]

    def _flexible_add_node(self, node: T) -> None:
        if not self.has_node(node):
            self._add_new_node(node)

    def _rigid_add_node(self, node: T) -> None:
        if self.has_node(node):
            raise ValueError(f"Node {node} already exists!")
        self._add_new_node(node)

    def _add_new_node(self, node: T) -> None:
        self.graph[node] = set()

    def _flexible_add_edge(self, source: T, destination: T) -> None:
        """Add an edge even if the nodes are not yet in the graph."""
        if not self.has_node(source):
            self._add_new_node(source)
        if not self.has_node(destination):
            self._add_new_node(destination)
        self._add_edge(source, destination)

    def _rigid_add_edge(self, source: T, destination: T) -> None:
        if not self.has_node(destination):
            raise ValueError(
                f"Destination {destination} does not exist as node in graph."
            )
        self._add_edge(source, destination)

    def _add_edge(self, source: T, destination: T) -> None:
        self.graph[source].add(destination)

    @classmethod
    def from_list_of_edges(cls, edges: Sequence[Edge]) -> DependencyGraph:
        graph = cls()
        for edge in edges:
            graph._flexible_add_edge(edge.source, edge.destination)
        return graph

    def is_cyclic(self) -> bool:
        """Check whether there are loops in the graph"""
        visited = set()
        recursion_stack = set()

        def depth_first_search(node: T) -> bool:
            visited.add(node)
            recursion_stack.add(node)

            for neighbor in self.graph.get(node, []):
                if neighbor not in visited:
                    if depth_first_search(neighbor):
                        return True
                elif neighbor in recursion_stack:
                    return True

            recursion_stack.remove(node)
            return False

        # Perform DFS for each node in the graph
        for node in self.graph:
            if node not in visited:
                if depth_first_search(node):
                    return True

        return False

    def get_ordered_nodes(self) -> List[T]:
        """Return the order in which nodes must be created"""
        assert not self.is_cyclic(), "Graph contains a cycle! Must be acyclic!"
        visited = set()
        ordering = []

        def depth_first_search(node: T) -> None:
            nonlocal visited, ordering
            visited.add(node)
            for neighbor in self.graph.get(node, []):
                if neighbor not in visited:
                    depth_first_search(neighbor)
            ordering.append(node)

        for node in self.graph:
            if node not in visited:
                depth_first_search(node)

        return ordering



from elabftwcontrol.download.api import (
    IngestConfiguration,
    IngestJob,
    ObjectTypes,
    OutputFormats,
    TableCellContentType,
    TableShapes,
    get_experiment_templates_as_df,
    get_experiments_as_df,
    get_item_types_as_df,
    get_items_as_df,
)

__all__ = [
    "IngestConfiguration",
    "IngestJob",
    "ObjectTypes",
    "OutputFormats",
    "TableCellContentType",
    "TableShapes",
    "get_item_types_as_df",
    "get_experiment_templates_as_df",
    "get_items_as_df",
    "get_experiments_as_df",
]



from typing import Optional

from elabftwcontrol._logging import logger
from elabftwcontrol.client import ElabftwApi


class GlobalConfig:
    """A class to make it to configure and store a global connection instead of being forced to always create and pass a client"""

    _client: Optional[ElabftwApi] = None

    @classmethod
    def get_client(
        cls,
    ) -> ElabftwApi:
        if cls._client is None:
            raise RuntimeError(
                "No connection has been configured. Please use the `elabftwcontrol.connect` "
                "function to set up a connection."
            )
        return cls._client

    @classmethod
    def configure_client(
        cls,
        profile: str = "default",
        host_url: Optional[str] = None,
        api_key: Optional[str] = None,
        verify_ssl: Optional[bool] = None,
    ) -> None:
        if host_url is not None:
            client = ElabftwApi.new(
                host_url=host_url,
                api_key=api_key,
                verify_ssl=verify_ssl,
            )
        else:
            logger.info(f"No host url configured, trying to use profile: {profile}")
            try:
                client = ElabftwApi.from_config_file(profile=profile)
            except KeyError:
                raise RuntimeError(
                    f"Profile `{profile}` is not configured. Use `elabftwctl config` "
                    "on the command line to set up a new profile."
                )
            except Exception as e:
                raise RuntimeError(f"Unexpected error: {e}")
        cls._client = client

    @classmethod
    def reset_client(
        cls,
    ) -> None:
        cls._client = None

    @classmethod
    def test_connection(cls) -> None:
        client = cls.get_client()
        info = client.info
        try:
            info.get_info()
        except Exception as e:
            raise RuntimeError(f"Could not establish connection to eLabFTW: {e}")
        logger.info(f"Succesfully connected to eLabFTW host: {client.host_name}")


def connect(
    profile: str = "default",
    host_url: Optional[str] = None,
    api_key: Optional[str] = None,
    verify_ssl: Optional[bool] = None,
    test_connection: bool = True,
) -> None:
    GlobalConfig.configure_client(
        profile=profile,
        host_url=host_url,
        api_key=api_key,
        verify_ssl=verify_ssl,
    )
    if test_connection:
        GlobalConfig.test_connection()



from datetime import datetime
from typing import Any

import pytest

from elabftwcontrol.core.metadata import (
    MetadataField,
    ParsedMetadataToMetadataFieldList,
    ParsedMetadataToPandasDtype,
    ParsedMetadataToSimpleDict,
    TableCellContentType,
)
from elabftwcontrol.core.models import (
    ConfigMetadata,
    FieldTypeEnum,
    GroupModel,
    MetadataModel,
    SingleFieldModel,
)
from elabftwcontrol.testing_utils import assert_dicts_equal


class TestMetadataField:
    @pytest.mark.parametrize(
        ("data", "expected"),
        (
            (
                SingleFieldModel(
                    type=FieldTypeEnum.text,
                    position=3,
                    value="5",
                    unit="kg",
                ),
                MetadataField(
                    name="test",
                    type="text",
                    group="test_group",
                    position=3,
                    value="5",
                    unit="kg",
                ),
            ),
            (
                SingleFieldModel(),
                MetadataField(
                    name="test",
                    type="text",
                    group="test_group",
                    position=-1,
                    value="",
                    unit="",
                ),
            ),
            (
                SingleFieldModel(
                    type=FieldTypeEnum.number,
                    position=None,
                    value=None,
                ),
                MetadataField(
                    name="test",
                    type="number",
                    group="test_group",
                    position=-1,
                    value="",
                ),
            ),
        ),
    )
    def test_from_parsed_field(
        self,
        data: SingleFieldModel,
        expected: MetadataField,
    ) -> None:
        assert (
            MetadataField.from_parsed_field(
                data=data,
                name="test",
                group="test_group",
            )
            == expected
        )


class TestParsedMetadataToMetadataFieldList:
    def test_full_metadata_with_groups(self) -> None:
        parsed_metadata = MetadataModel(
            elabftw=ConfigMetadata(
                extra_fields_groups=[
                    GroupModel(id=1, name="group 1"),
                    GroupModel(id=2, name="group 2"),
                ]
            ),
            extra_fields={
                "field 1": SingleFieldModel(
                    type=FieldTypeEnum("number"),
                    unit="h",
                    units=["h"],
                    group_id=2,
                    position=8,
                    description="something",
                ),
                "field 2": SingleFieldModel(
                    type=FieldTypeEnum("date"),
                    value="x",
                    group_id=1,
                    position=None,
                ),
                "field 3": SingleFieldModel(
                    value="anything",
                    group_id=None,
                    position=10,
                ),
            },
        )
        fields = ParsedMetadataToMetadataFieldList()(parsed_metadata)
        expected: list[MetadataField] = [
            MetadataField(
                name="field 1",
                type="number",
                group="group 2",
                position=8,
                description="something",
                value="",
                unit="h",
            ),
            MetadataField(
                name="field 2",
                type="date",
                group="group 1",
                position=-1,
                value="x",
            ),
            MetadataField(
                name="field 3",
                type="text",
                position=10,
                value="anything",
            ),
        ]
        assert len(fields) == len(expected)
        assert set(fields) == set(expected)

    def test_full_metadata_no_groups(self) -> None:
        parsed_metadata = MetadataModel(
            extra_fields={
                "field 1": SingleFieldModel(
                    type=FieldTypeEnum("number"),
                    unit="h",
                    units=["h"],
                    group_id=2,
                    position=8,
                    description="something",
                ),
                "field 2": SingleFieldModel(
                    type=FieldTypeEnum("date"),
                    value="x",
                    group_id=1,
                    position=None,
                ),
                "field 3": SingleFieldModel(
                    value="anything",
                    group_id=None,
                    position=10,
                ),
            }
        )
        fields = ParsedMetadataToMetadataFieldList()(parsed_metadata)
        expected: list[MetadataField] = [
            MetadataField(
                name="field 1",
                type="number",
                description="something",
                position=8,
                unit="h",
            ),
            MetadataField(
                name="field 2",
                type="date",
                position=-1,
                value="x",
            ),
            MetadataField(
                name="field 3",
                type="text",
                position=10,
                value="anything",
            ),
        ]
        assert len(fields) == len(expected)
        assert set(fields) == set(expected)


class TestParsedMetadataToSimpleDict:
    metadata = MetadataModel(
        extra_fields={
            "field 3": SingleFieldModel(
                position=10,
                value="20",
                type=FieldTypeEnum("number"),
                unit="kg",
            ),
            "field 1": SingleFieldModel(
                position=8,
                type=FieldTypeEnum("number"),
                unit="kg",
            ),
            "field 2": SingleFieldModel(
                value="20",
                type=FieldTypeEnum("number"),
            ),
            "field 6": SingleFieldModel(
                position=13,
                value="20",
                unit="kg",
            ),
            "field 5": SingleFieldModel(
                position=12,
                value="1999-09-09T19:19",
                type=FieldTypeEnum("datetime-local"),
                unit="kg",
            ),
            "field 4": SingleFieldModel(
                position=11,
                value="notanumber",
                type=FieldTypeEnum("number"),
                unit="kg",
            ),
            "field 7": SingleFieldModel(
                position=99,
                value="blabla",
                type=FieldTypeEnum("text"),
            ),
        },
    )

    @pytest.mark.parametrize(
        ("cell_content", "expected"),
        (
            (
                "value",
                {
                    "field 2": 20,
                    "field 1": float("nan"),
                    "field 3": 20,
                    "field 4": float("nan"),
                    "field 5": datetime(1999, 9, 9, 19, 19),
                    "field 6": "20",
                    "field 7": "blabla",
                },
            ),
            (
                "unit",
                {
                    "field 2": "",
                    "field 1": "",
                    "field 3": "kg",
                    "field 4": "kg",
                    "field 5": "kg",
                    "field 6": "kg",
                    "field 7": "",
                },
            ),
            (
                "combined",
                {
                    "field 2": "20",
                    "field 1": "",
                    "field 3": "20 kg",
                    "field 4": "notanumber kg",
                    "field 5": "1999-09-09T19:19 kg",
                    "field 6": "20 kg",
                    "field 7": "blabla",
                },
            ),
        ),
    )
    def test_get_simple_dict(
        self,
        cell_content: TableCellContentType,
        expected: dict[str, Any],
    ) -> None:
        transformer = ParsedMetadataToSimpleDict.new(
            cell_content=cell_content,
        )
        transformed = transformer(self.metadata)
        assert_dicts_equal(transformed, expected, order_is_important=True)


class TestParsedMetadataToPandasDtype:
    metadata = MetadataModel(
        extra_fields={
            "field 1": SingleFieldModel(
                type=FieldTypeEnum("number"),
            ),
            "field 2": SingleFieldModel(
                type=FieldTypeEnum("text"),
            ),
            "field 3": SingleFieldModel(
                type=FieldTypeEnum("time"),
            ),
            "field 4": SingleFieldModel(
                type=FieldTypeEnum("experiments"),
            ),
        },
    )

    @pytest.mark.parametrize(
        ("cell_content", "expected"),
        (
            (
                "value",
                {
                    "field 1": "Float64",
                    "field 2": "string",
                    "field 3": "datetime64[ns]",
                    "field 4": "Int64",
                },
            ),
            (
                "unit",
                {
                    "field 1": "string",
                    "field 2": "string",
                    "field 3": "string",
                    "field 4": "string",
                },
            ),
            (
                "combined",
                {
                    "field 1": "string",
                    "field 2": "string",
                    "field 3": "datetime64[ns]",
                    "field 4": "Int64",
                },
            ),
        ),
    )
    def test_get_pandas_dtype(
        self,
        cell_content: TableCellContentType,
        expected: dict[str, Any],
    ) -> None:
        transformer = ParsedMetadataToPandasDtype.new(
            cell_content=cell_content,
        )
        transformed = transformer(self.metadata)
        assert transformed == expected



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, Iterator, TypeVar

from typing_extensions import Self

from elabftwcontrol._logging import logger
from elabftwcontrol.client import ElabftwApi
from elabftwcontrol.core.manifests import ManifestIndex, Patch, _Spec
from elabftwcontrol.core.models import IdNode, NameNode
from elabftwcontrol.types import ElabObjectType
from elabftwcontrol.upload.diff import Diff
from elabftwcontrol.upload.state import EnrichedObj, State


class JobType(Enum):
    CREATE = "create"
    UPDATE = "update"
    DELETE = "delete"


T = TypeVar("T")


@dataclass
class WorkEvaluatorState:
    """Keeps track of the objects in eLabFTW during the job.
    Provides easy translation between manifest name nodes and elab state id nodes.
    """

    manifest_index: ManifestIndex
    elab_state: State
    name_to_id: dict[NameNode, IdNode]
    id_to_name: dict[IdNode, NameNode]

    def get_id(self, node: NameNode) -> IdNode | None:
        return self.name_to_id.get(node)

    def get_name(self, node: IdNode) -> NameNode | None:
        return self.id_to_name.get(node)

    def get_state_obj(self, node: IdNode) -> EnrichedObj | None:
        return self.elab_state.get(node)

    def get_manifest(self, node: NameNode) -> _Spec:
        return self.manifest_index[node]

    def contains_id(self, node: IdNode) -> bool:
        return node in self.id_to_name

    def contains_name(self, node: NameNode) -> bool:
        return node in self.name_to_id

    def add(self, name: NameNode, id: IdNode) -> None:
        self.name_to_id[name] = id
        self.id_to_name[id] = name

    def remove_by_name(self, name: NameNode) -> None:
        id = self.name_to_id[name]
        del self.id_to_name[id]
        del self.name_to_id[name]

    def remove_by_id(self, id: IdNode) -> None:
        name = self.id_to_name[id]
        del self.name_to_id[name]
        del self.id_to_name[id]

    def get_parent_node_id(self, name: NameNode) -> IdNode | None:
        parent_name = self.manifest_index.parents.get(name)
        if parent_name is None:
            return None
        return self.get_id(parent_name)

    @classmethod
    def new(
        cls,
        elab_state: State,
        manifest_index: ManifestIndex,
    ) -> Self:
        name_to_id = {}
        id_to_name = {}
        for type_and_id, enriched_obj in elab_state.items():
            type_and_name = NameNode(enriched_obj.type, enriched_obj.name)
            name_to_id[type_and_name] = type_and_id
            id_to_name[type_and_id] = type_and_name
        return cls(
            manifest_index=manifest_index,
            elab_state=elab_state,
            name_to_id=name_to_id,
            id_to_name=id_to_name,
        )


class WorkTypes(str, Enum):
    APPLY = "apply"
    DESTROY = "destroy"


@dataclass
class WorkEvaluator:
    """Compares state and manifests and calculates the work plan to be done"""

    job_state: WorkEvaluatorState
    version: str | None

    @classmethod
    def new(
        cls,
        manifest_index: ManifestIndex,
        elab_state: State,
        version: str | None,
    ) -> Self:
        job_state = WorkEvaluatorState.new(
            elab_state=elab_state,
            manifest_index=manifest_index,
        )
        return cls(job_state=job_state, version=version)

    @property
    def manifest_index(self) -> ManifestIndex:
        return self.job_state.manifest_index

    @property
    def elab_state(self) -> State:
        return self.job_state.elab_state

    def evaluate_apply(self) -> WorkPlan:
        nodes_to_apply = self.manifest_index.get_node_creation_order()
        plan = WorkPlan.new()

        for name_node in nodes_to_apply:
            if not self.job_state.contains_name(name_node):
                new_task = ElabOperation.new_create_obj(
                    name_node=name_node,
                    job_state=self.job_state,
                )
                plan.add_task(new_task)
                new_task = ElabOperation.new_update_obj(
                    name_node=name_node,
                    patch=self.get_patch(name_node),
                    job_state=self.job_state,
                )
                plan.add_task(new_task)

            else:
                patch = self.get_patch(name_node)
                if patch:
                    new_task = ElabOperation.new_update_obj(
                        name_node=name_node,
                        patch=patch,
                        job_state=self.job_state,
                    )
                    plan.add_task(new_task)

        return plan

    def get_patch(
        self,
        name_node: NameNode,
    ) -> Patch:
        spec = self.manifest_index[name_node]
        return spec.to_patch(
            name_node=name_node,
            state=self.job_state,
            version=self.version,
        )

    def evaluate_destroy(self) -> WorkPlan:
        nodes_to_delete = self.manifest_index.get_node_deletion_order()
        plan = WorkPlan.new()

        for name_node in nodes_to_delete:
            if self.job_state.contains_name(name_node):
                new_task = ElabOperation.new_delete_obj(
                    name_node=name_node,
                    job_state=self.job_state,
                )
                plan.add_task(new_task)
            else:
                logger.warning(
                    f"The object {name_node} does not exist. Can not delete."
                )

        return plan


@dataclass
class WorkPlan:
    tasks: list[ElabOperation]

    def __bool__(self) -> bool:
        return len(self.tasks) > 0

    def _to_str_iter(self) -> Iterator[str]:
        for task in self.tasks:
            yield str(task)

    def print_to_console(
        self,
        print_fn: Callable[[Any], None] = print,
    ) -> None:
        for text in self._to_str_iter():
            print_fn(text)

    @property
    def n_tasks(self) -> int:
        return len(self.tasks)

    @classmethod
    def new(cls) -> Self:
        return cls([])

    def add_task(self, task: ElabOperation) -> None:
        self.tasks.append(task)

    def execute(
        self,
        client: ElabftwApi,
        start: int = 0,
        n_retries: int = 5,
    ) -> None:
        for i, task in enumerate(self.tasks[start:], start=start):
            error: Exception
            for retry in range(n_retries):
                try:
                    logger.info(
                        f"Attempting task ({i+1}/{self.n_tasks}), try ({retry+1}/{n_retries})"
                    )
                    task(client)
                    break
                except Exception as e:
                    error = e
            else:
                logger.error(f"Failed on task {i}")
                raise error
        logger.info(f"Completed the work plan of {self.n_tasks}")


class ElabOperationError(Exception): ...


class DeletionError(ElabOperationError): ...


class CreationError(ElabOperationError): ...


class PatchingError(ElabOperationError): ...


@dataclass(frozen=True)
class ElabOperation:
    name_node: NameNode
    action: Callable[[ElabftwApi], IdNode]
    success_callback: Callable[[IdNode], None] | None
    failure_callback: Callable[[NameNode, Exception], None] | None

    def __str__(self) -> str:
        return str(self.action)

    def __call__(self, api: ElabftwApi) -> None:
        try:
            id_node = self.action(api)
        except Exception as e:
            logger.error("Could not execute '%s': %s" % (str(self.action), e))
            if self.failure_callback is not None:
                self.failure_callback(self.name_node, e)
            return

        if self.success_callback is not None:
            self.success_callback(id_node)

    @classmethod
    def new_create_obj(
        cls,
        name_node: NameNode,
        job_state: WorkEvaluatorState,
    ) -> Self:
        obj_type = name_node.kind

        def create_method_factory(api: ElabftwApi) -> CreateMethod:
            match obj_type:
                case ElabObjectType.ITEMS_TYPE:
                    return api.items_types.create
                case ElabObjectType.EXPERIMENTS_TEMPLATE:
                    return api.experiments_templates.create
                case ElabObjectType.EXPERIMENT:
                    return api.experiments.create
                case ElabObjectType.ITEM:

                    def create_item() -> int:
                        parent_node = job_state.get_parent_node_id(name_node)
                        if parent_node is None:
                            raise RuntimeError(
                                f"{name_node} requires a an item type id, which was not found"
                            )
                        category_id = parent_node.id
                        return api.items.create(category_id=category_id)

                    return create_item
                case _:
                    raise ValueError

        def success_callback(id_node: IdNode) -> None:
            pass

        def failure_callback(node: NameNode, error: Exception) -> None:
            raise CreationError(f"{node} failed to be created: {error}")

        action = CreateObj(
            name_node=name_node,
            job_state=job_state,
            create_method_factory=create_method_factory,
        )

        return cls(
            name_node=name_node,
            action=action,
            success_callback=success_callback,
            failure_callback=failure_callback,
        )

    @classmethod
    def new_update_obj(
        cls,
        name_node: NameNode,
        patch: Patch,
        job_state: WorkEvaluatorState,
    ) -> Self:
        action = UpdateObj(
            name_node=name_node,
            job_state=job_state,
            patch=patch,
        )

        def success_callback(id_node: IdNode) -> None:
            pass

        def failure_callback(node: NameNode, error: Exception) -> None:
            raise PatchingError(f"{node} failed to be patched: {error}")

        return cls(
            name_node=name_node,
            action=action,
            success_callback=success_callback,
            failure_callback=failure_callback,
        )

    @classmethod
    def new_delete_obj(
        cls,
        name_node: NameNode,
        job_state: WorkEvaluatorState,
    ) -> Self:
        obj_type = name_node.kind

        def delete_method_factory(api: ElabftwApi) -> DeleteMethod:
            match obj_type:
                case ElabObjectType.ITEMS_TYPE:
                    return api.items_types.delete
                case ElabObjectType.EXPERIMENTS_TEMPLATE:
                    return api.experiments_templates.delete
                case ElabObjectType.EXPERIMENT:
                    return api.experiments.delete
                case ElabObjectType.ITEM:
                    return api.items.delete
                case _:
                    raise ValueError

        action = DeleteObj(
            name_node=name_node,
            job_state=job_state,
            delete_method_factory=delete_method_factory,
        )

        def success_callback(id_node: IdNode) -> None:
            pass

        def failure_callback(node: NameNode, error: Exception) -> None:
            raise RuntimeError(f"{node} failed to be deleted: {error}")

        return cls(
            name_node=name_node,
            action=action,
            success_callback=success_callback,
            failure_callback=failure_callback,
        )


_RED = "\033[31m"
_GREEN = "\033[32m"
_YELLOW = "\033[33m"
_RESET = "\033[0m"


CreateMethod = Callable[[], int]


class CreateObj:
    """Command for creating a new object and returning an id"""

    def __init__(
        self,
        name_node: NameNode,
        job_state: WorkEvaluatorState,
        create_method_factory: Callable[[ElabftwApi], CreateMethod],
    ) -> None:
        self.name_node = name_node
        self.job_state = job_state
        self.create_method_factory = create_method_factory

    def __str__(self) -> str:
        return f"{_GREEN}+{_RESET} Creation of new {self.name_node}"

    def __call__(self, api: ElabftwApi) -> IdNode:
        create_method = self.create_method_factory(api)
        id = create_method()
        new_node = IdNode(kind=self.name_node.kind, id=id)
        self.job_state.add(name=self.name_node, id=new_node)
        return new_node


UpdateMethod = Callable[[int, dict[str, Any]], None]


class UpdateObj:
    def __init__(
        self,
        name_node: NameNode,
        patch: Patch,
        job_state: WorkEvaluatorState,
    ) -> None:
        self.name_node = name_node
        self.patch = patch
        self.job_state = job_state

    def __call__(self, api: ElabftwApi) -> IdNode:
        if self.id_node is None:
            raise RuntimeError(f"{self.name_node} does not exist.")

        self.patch(api)
        return self.id_node

    def __str__(self) -> str:
        if self.id_node is None:
            id_text = "ID unknown"
        else:
            id_text = str(self.id_node.id)

        return f"""\
{_YELLOW}~{_RESET} Patching of {self.name_node} ({id_text})

{self.diff.to_str(indent=4)}
"""

    @property
    def diff(self) -> Diff:
        return self.patch.diff

    @property
    def id_node(self) -> IdNode | None:
        return self.job_state.get_id(self.name_node)


DeleteMethod = Callable[[int], None]


class DeleteObj:
    """Command for deleting an API object and updating the state"""

    def __init__(
        self,
        name_node: NameNode,
        job_state: WorkEvaluatorState,
        delete_method_factory: Callable[[ElabftwApi], DeleteMethod],
    ) -> None:
        self.job_state = job_state
        self.name_node = name_node
        self.delete_method_factory = delete_method_factory

    def __call__(self, api: ElabftwApi) -> IdNode:
        id_node = self.id_node
        if id_node is None:
            raise RuntimeError(f"{self.name_node} does not exist.")

        delete_method = self.delete_method_factory(api)
        delete_method(id_node.id)
        self.job_state.remove_by_id(id_node)
        return id_node

    def __str__(self) -> str:
        if self.id_node is not None:
            node_id = str(self.id_node.id)
        else:
            node_id = "ID unknown"
        return f"{_RED}-{_RESET} Deletion of {self.name_node} ({node_id})"

    @property
    def id_node(self) -> IdNode | None:
        return self.job_state.get_id(self.name_node)
